import dataclasses
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import plot_error_bias_colored_scatter


@dataclasses.dataclass
class RegressionTopErrorMetricResults:
    curr_mean_err_per_group: Dict[str, Dict[str, float]]
    curr_scatter: Dict[str, Dict[str, pd.Series]]
    ref_mean_err_per_group: Optional[Dict[str, Dict[str, float]]]
    ref_scatter: Optional[Dict[str, Dict[str, pd.Series]]]


class RegressionTopErrorMetric(Metric[RegressionTopErrorMetricResults]):
    def calculate(self, data: InputData) -> RegressionTopErrorMetricResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        curr_df = data.current_data
        ref_df = data.reference_data
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        if not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")
        curr_df = self._make_df_for_plot(curr_df, target_name, prediction_name, None)
        curr_error = curr_df[prediction_name] - curr_df[target_name]
        quantile_5 = np.quantile(curr_error, 0.05)
        quantile_95 = np.quantile(curr_error, 0.95)

        curr_df["Error bias"] = list(
            map(
                lambda x: "Underestimation" if x <= quantile_5 else "Majority" if x < quantile_95 else "Overestimation",
                curr_error,
            )
        )
        curr_scatter = self._get_data_for_scatter(curr_df, target_name, prediction_name)
        curr_mean_err_per_group = self._calculate_underperformance(curr_error, quantile_5, quantile_95)

        ref_scatter = None
        ref_mean_err_per_group = None
        if ref_df is not None:
            ref_df = self._make_df_for_plot(ref_df, target_name, prediction_name, None)
            ref_error = ref_df[prediction_name] - ref_df[target_name]
            quantile_5 = np.quantile(ref_error, 0.05)
            quantile_95 = np.quantile(ref_error, 0.95)

            ref_df["Error bias"] = list(
                map(
                    lambda x: "Underestimation"
                    if x <= quantile_5
                    else "Majority"
                    if x < quantile_95
                    else "Overestimation",
                    ref_error,
                )
            )
            ref_scatter = self._get_data_for_scatter(ref_df, target_name, prediction_name)
            ref_mean_err_per_group = self._calculate_underperformance(ref_error, quantile_5, quantile_95)
        return RegressionTopErrorMetricResults(
            curr_mean_err_per_group=curr_mean_err_per_group,
            curr_scatter=curr_scatter,
            ref_mean_err_per_group=ref_mean_err_per_group,
            ref_scatter=ref_scatter,
        )

    def _make_df_for_plot(self, df, target_name: str, prediction_name: str, datetime_column_name: Optional[str]):
        result = df.replace([np.inf, -np.inf], np.nan)
        if datetime_column_name is not None:
            result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name, datetime_column_name])
            return result.sort_values(datetime_column_name)
        result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name])
        return result.sort_index()

    def _get_data_for_scatter(self, df: pd.DataFrame, target_name: str, prediction_name: str):
        scatter = {}
        scatter["Underestimation"] = {
            "Predicted value": df.loc[df["Error bias"] == "Underestimation", prediction_name],
            "Actual value": df.loc[df["Error bias"] == "Underestimation", target_name],
        }
        scatter["Majority"] = {
            "Predicted value": df.loc[df["Error bias"] == "Majority", prediction_name],
            "Actual value": df.loc[df["Error bias"] == "Majority", target_name],
        }
        scatter["Overestimation"] = {
            "Predicted value": df.loc[df["Error bias"] == "Overestimation", prediction_name],
            "Actual value": df.loc[df["Error bias"] == "Overestimation", target_name],
        }
        return scatter

    def _calculate_underperformance(
        self, error: pd.Series, quantile_5: float, quantile_95: float, conf_interval_n_sigmas: int = 1
    ):

        mae_under = np.mean(error[error <= quantile_5])
        mae_exp = np.mean(error[(error > quantile_5) & (error < quantile_95)])
        mae_over = np.mean(error[error >= quantile_95])

        sd_under = np.std(error[error <= quantile_5], ddof=1)
        sd_exp = np.std(error[(error > quantile_5) & (error < quantile_95)], ddof=1)
        sd_over = np.std(error[error >= quantile_95], ddof=1)

        return {
            "majority": {"mean_error": float(mae_exp), "std_error": conf_interval_n_sigmas * float(sd_exp)},
            "underestimation": {"mean_error": float(mae_under), "std_error": conf_interval_n_sigmas * float(sd_under)},
            "overestimation": {"mean_error": float(mae_over), "std_error": conf_interval_n_sigmas * float(sd_over)},
        }


@default_renderer(wrap_type=RegressionTopErrorMetric)
class RegressionTopErrorMetricRenderer(MetricRenderer):
    def render_json(self, obj: RegressionTopErrorMetric) -> dict:
        return {}

    def render_html(self, obj: RegressionTopErrorMetric) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        curr_mean_err_per_group = result.curr_mean_err_per_group
        curr_scatter = result.curr_scatter
        ref_mean_err_per_group = result.ref_mean_err_per_group
        ref_scatter = result.ref_scatter

        res = [
            header_text(label="Error Bias Table"),
            counter(
                title="Current: Mean Error per Group (+/- std)",
                counters=[
                    CounterData("Majority(90%)", self._format_value(curr_mean_err_per_group, "majority")),
                    CounterData("Underestimation(5%)", self._format_value(curr_mean_err_per_group, "underestimation")),
                    CounterData("Overestimation(5%)", self._format_value(curr_mean_err_per_group, "overestimation")),
                ],
            ),
        ]
        if ref_mean_err_per_group is not None:
            res.append(
                counter(
                    title="Reference: Mean Error per Group (+/- std)",
                    counters=[
                        CounterData("Majority(90%)", self._format_value(ref_mean_err_per_group, "majority")),
                        CounterData(
                            "Underestimation(5%)", self._format_value(ref_mean_err_per_group, "underestimation")
                        ),
                        CounterData("Overestimation(5%)", self._format_value(ref_mean_err_per_group, "overestimation")),
                    ],
                )
            )
        res.append(header_text(label="Predicted vs Actual per Group"))
        fig = plot_error_bias_colored_scatter(curr_scatter, ref_scatter, color_options=self.color_options)

        res.append(
            BaseWidgetInfo(
                title="",
                size=2,
                type="big_graph",
                params={"data": fig["data"], "layout": fig["layout"]},
            )
        )
        return res

    def _format_value(self, result, counter_type):
        return f"{round(result[counter_type]['mean_error'], 2)}" f" ({round(result[counter_type]['std_error'], 2)})"



import dataclasses
import json
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.classification_performance import PredictionData
from evidently.calculations.classification_performance import get_prediction_data
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import DataDefinition


@dataclasses.dataclass
class ClassificationQualityByFeatureTableResults:
    current_plot_data: pd.DataFrame
    reference_plot_data: Optional[pd.DataFrame]
    target_name: str
    curr_predictions: PredictionData
    ref_predictions: Optional[PredictionData]
    columns: List[str]


class ClassificationQualityByFeatureTable(Metric[ClassificationQualityByFeatureTableResults]):
    columns: Optional[List[str]]
    text_features_gen: Optional[
        Dict[str, Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]]
    ]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns
        self.text_features_gen = None

    def required_features(self, data_definition: DataDefinition):
        if len(data_definition.get_columns("text_features")) > 0:
            text_cols = [col.column_name for col in data_definition.get_columns("text_features")]
            text_features_gen = {}
            text_features_gen_result = []
            for col in text_cols:
                col_dict: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]] = {}
                col_dict[f"{col}: Text Length"] = TextLength(col)
                col_dict[f"{col}: Non Letter Character %"] = NonLetterCharacterPercentage(col)
                col_dict[f"{col}: OOV %"] = OOVWordsPercentage(col)

                text_features_gen_result += [
                    col_dict[f"{col}: Text Length"],
                    col_dict[f"{col}: Non Letter Character %"],
                    col_dict[f"{col}: OOV %"],
                ]
                text_features_gen[col] = col_dict
            self.text_features_gen = text_features_gen

            return text_features_gen_result
        else:
            return []

    def get_parameters(self) -> tuple:
        return ()

    def calculate(self, data: InputData) -> ClassificationQualityByFeatureTableResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        curr_df = data.current_data.copy()
        ref_df = None
        if data.reference_data is not None:
            ref_df = data.reference_data.copy()
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' should be present")
        curr_predictions = get_prediction_data(data.current_data, dataset_columns, data.column_mapping.pos_label)
        ref_predictions = None
        if ref_df is not None:
            ref_predictions = get_prediction_data(data.reference_data, dataset_columns, data.column_mapping.pos_label)
        if self.columns is None:
            columns = (
                dataset_columns.num_feature_names
                + dataset_columns.cat_feature_names
                + dataset_columns.text_feature_names
            )
        else:
            columns = list(
                np.intersect1d(
                    self.columns,
                    (
                        dataset_columns.num_feature_names
                        + dataset_columns.cat_feature_names
                        + dataset_columns.text_feature_names
                    ),
                )
            )
        # process text columns
        if (
            self.text_features_gen is not None
            and len(np.intersect1d(list(self.text_features_gen.keys()), columns)) >= 1
        ):
            for col in np.intersect1d(list(self.text_features_gen.keys()), columns):
                columns += list(self.text_features_gen[col].keys())
                columns.remove(col)
                curr_text_df = pd.concat(
                    [data.get_current_column(x.feature_name()) for x in list(self.text_features_gen[col].values())],
                    axis=1,
                )
                curr_text_df.columns = list(self.text_features_gen[col].keys())
                curr_df = pd.concat([curr_df.reset_index(drop=True), curr_text_df.reset_index(drop=True)], axis=1)

                if ref_df is not None:
                    ref_text_df = pd.concat(
                        [
                            data.get_reference_column(x.feature_name())
                            for x in list(self.text_features_gen[col].values())
                        ],
                        axis=1,
                    )
                    ref_text_df.columns = list(self.text_features_gen[col].keys())
                    ref_df = pd.concat([ref_df.reset_index(drop=True), ref_text_df.reset_index(drop=True)], axis=1)

        return ClassificationQualityByFeatureTableResults(
            current_plot_data=curr_df,
            reference_plot_data=ref_df,
            curr_predictions=curr_predictions,
            ref_predictions=ref_predictions,
            columns=columns,
            target_name=target_name,
        )


@default_renderer(wrap_type=ClassificationQualityByFeatureTable)
class ClassificationQualityByFeatureTableRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationQualityByFeatureTable) -> dict:
        return {}

    def render_html(self, obj: ClassificationQualityByFeatureTable) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_data = result.current_plot_data
        reference_data = result.reference_plot_data
        target_name = result.target_name
        curr_predictions = result.curr_predictions
        ref_predictions = result.ref_predictions
        columns = result.columns
        labels = curr_predictions.labels

        color_options = self.color_options

        current_data["prediction_labels"] = curr_predictions.predictions.values

        if reference_data is not None and ref_predictions is not None:
            reference_data["prediction_labels"] = ref_predictions.predictions.values

        additional_graphs_data = []
        params_data = []
        for feature_name in columns:
            # add data for table in params

            params_data.append(
                {
                    "details": {
                        "parts": [{"title": "All", "id": "All" + "_" + str(feature_name)}]
                        + [{"title": str(label), "id": feature_name + "_" + str(label)} for label in labels],
                        "insights": [],
                    },
                    "f1": feature_name,
                }
            )

            # create confusion based plots
            current_data["dataset"] = "Current"
            merged_data = current_data
            facet_col: Optional[str] = None
            category_orders: Optional[dict] = None
            if reference_data is not None:
                reference_data["dataset"] = "Reference"
                merged_data = pd.concat([reference_data, current_data])
                facet_col = "dataset"
                category_orders = {"dataset": ["Current", "Reference"]}

            fig = px.histogram(
                merged_data,
                x=feature_name,
                color=target_name,
                facet_col=facet_col,
                histnorm="",
                barmode="overlay",
                category_orders=category_orders,
            )

            fig_json = json.loads(fig.to_json())

            # write plot data in table as additional data
            additional_graphs_data.append(
                AdditionalGraphInfo(
                    "All" + "_" + str(feature_name),
                    {"data": fig_json["data"], "layout": fig_json["layout"]},
                )
            )

            # Probas plots
            if curr_predictions.prediction_probas is not None:
                ref_columns = columns + ["prediction_labels", target_name]
                current_data = pd.concat([current_data[ref_columns], curr_predictions.prediction_probas], axis=1)
                if (
                    reference_data is not None
                    and ref_predictions is not None
                    and ref_predictions.prediction_probas is not None
                ):
                    reference_data = pd.concat([reference_data[ref_columns], ref_predictions.prediction_probas], axis=1)

                if reference_data is not None:
                    cols = 2
                    subplot_titles = ["current", "reference"]
                else:
                    cols = 1
                    subplot_titles = [""]

                for label in labels:
                    fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)

                    # current Prediction
                    fig.add_trace(
                        go.Scatter(
                            x=current_data[current_data[target_name] == label][feature_name],
                            y=current_data[current_data[target_name] == label][label],
                            mode="markers",
                            name=str(label),
                            legendgroup=str(label),
                            marker=dict(
                                size=6,
                                # set color equal to a variable
                                color=color_options.get_current_data_color(),
                            ),
                        ),
                        row=1,
                        col=1,
                    )

                    fig.add_trace(
                        go.Scatter(
                            x=current_data[current_data[target_name] != label][feature_name],
                            y=current_data[current_data[target_name] != label][label],
                            mode="markers",
                            name="other",
                            legendgroup="other",
                            marker=dict(
                                size=6,
                                # set color equal to a variable
                                color=color_options.get_reference_data_color(),
                            ),
                        ),
                        row=1,
                        col=1,
                    )

                    fig.update_xaxes(title_text=feature_name, showgrid=True, row=1, col=1)

                    # REF
                    if reference_data is not None:
                        fig.add_trace(
                            go.Scatter(
                                x=reference_data[reference_data[target_name] == label][feature_name],
                                y=reference_data[reference_data[target_name] == label][label],
                                mode="markers",
                                name=str(label),
                                legendgroup=str(label),
                                showlegend=False,
                                marker=dict(size=6, color=color_options.get_current_data_color()),
                            ),
                            row=1,
                            col=2,
                        )

                        fig.add_trace(
                            go.Scatter(
                                x=reference_data[reference_data[target_name] != label][feature_name],
                                y=reference_data[reference_data[target_name] != label][label],
                                mode="markers",
                                name="other",
                                legendgroup="other",
                                showlegend=False,
                                marker=dict(size=6, color=color_options.get_reference_data_color()),
                            ),
                            row=1,
                            col=2,
                        )
                        fig.update_xaxes(title_text=feature_name, row=1, col=2)

                    fig.update_layout(
                        yaxis_title="Probability",
                        yaxis=dict(range=(-0.1, 1.1), showticklabels=True),
                    )

                    fig_json = json.loads(fig.to_json())

                    # write plot data in table as additional data
                    additional_graphs_data.append(
                        AdditionalGraphInfo(
                            feature_name + "_" + str(label),
                            {"data": fig_json["data"], "layout": fig_json["layout"]},
                        )
                    )
            # labels plots
            else:
                for label in labels:

                    def confusion_func(row, label=label):
                        return self._confusion(row, target_name, "prediction_labels", label)

                    merged_data["Confusion"] = merged_data.apply(confusion_func, axis=1)

                    fig = px.histogram(
                        merged_data,
                        x=feature_name,
                        color="Confusion",
                        facet_col=facet_col,
                        histnorm="",
                        barmode="overlay",
                        category_orders=category_orders,
                    )
                    fig_json = json.loads(fig.to_json())

                    # write plot data in table as additional data
                    additional_graphs_data.append(
                        AdditionalGraphInfo(
                            feature_name + "_" + str(label),
                            {"data": fig_json["data"], "layout": fig_json["layout"]},
                        )
                    )
        return [
            header_text(label="Classification Quality By Feature"),
            BaseWidgetInfo(
                title="",
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": min(len(columns), 10),
                    "columns": [{"title": "Feature", "field": "f1"}],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            ),
        ]

    def _confusion(self, row, target_column, prediction_column, label):
        if row[target_column] == label and row[prediction_column] == label:
            return "TP"
        if row[target_column] != label and row[prediction_column] == label:
            return "FP"
        if row[target_column] == label and row[prediction_column] != label:
            return "FN"
        return "TN"  # last option



#!/usr/bin/env python
# coding: utf-8

from .cat_target_drift_tab import CatTargetDriftTab
from .classification_performance_tab import ClassificationPerformanceTab
from .data_drift_tab import DataDriftTab
from .data_quality_tab import DataQualityTab
from .num_target_drift_tab import NumTargetDriftTab
from .prob_classification_performance_tab import ProbClassificationPerformanceTab
from .regression_performance_tab import RegressionPerformanceTab



import warnings

import evidently.dashboard.widgets
from evidently.dashboard.widgets import *

__path__ = evidently.dashboard.widgets.__path__  # type: ignore

warnings.warn(
    "'import evidently.widgets' is deprecated, use 'import evidently.dashboard.widgets'"
)



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class RegColoredPredActualWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [RegressionPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = RegressionPerformanceAnalyzer.get_results(analyzers_results)
        results_utility_columns = results.columns.utility_columns

        if (
            results_utility_columns.target is None
            or results_utility_columns.prediction is None
        ):
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )

            return None

        if self.dataset == "current":
            dataset_to_plot = (
                current_data.copy(deep=False) if current_data is not None else None
            )

        else:
            dataset_to_plot = reference_data.copy(deep=False)

        if dataset_to_plot is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires reference dataset but it is None"
                )
            return None

        dataset_to_plot.replace([np.inf, -np.inf], np.nan, inplace=True)
        dataset_to_plot.dropna(
            axis=0,
            how="any",
            inplace=True,
            subset=[results_utility_columns.target, results_utility_columns.prediction],
        )

        error = (
            dataset_to_plot[results_utility_columns.prediction]
            - dataset_to_plot[results_utility_columns.target]
        )

        quantile_5 = np.quantile(error, 0.05)
        quantile_95 = np.quantile(error, 0.95)

        dataset_to_plot["Error bias"] = list(
            map(
                lambda x: "Underestimation"
                if x <= quantile_5
                else "Majority"
                if x < quantile_95
                else "Overestimation",
                error,
            )
        )

        # plot output correlations
        pred_actual = go.Figure()

        pred_actual.add_trace(
            go.Scatter(
                x=dataset_to_plot[dataset_to_plot["Error bias"] == "Underestimation"][
                    results_utility_columns.target
                ],
                y=dataset_to_plot[dataset_to_plot["Error bias"] == "Underestimation"][
                    results_utility_columns.prediction
                ],
                mode="markers",
                name="Underestimation",
                marker=dict(color=color_options.underestimation_color, showscale=False),
            )
        )

        pred_actual.add_trace(
            go.Scatter(
                x=dataset_to_plot[dataset_to_plot["Error bias"] == "Overestimation"][
                    results_utility_columns.target
                ],
                y=dataset_to_plot[dataset_to_plot["Error bias"] == "Overestimation"][
                    results_utility_columns.prediction
                ],
                mode="markers",
                name="Overestimation",
                marker=dict(color=color_options.overestimation_color, showscale=False),
            )
        )

        pred_actual.add_trace(
            go.Scatter(
                x=dataset_to_plot[dataset_to_plot["Error bias"] == "Majority"][
                    results_utility_columns.target
                ],
                y=dataset_to_plot[dataset_to_plot["Error bias"] == "Majority"][
                    results_utility_columns.prediction
                ],
                mode="markers",
                name="Majority",
                marker=dict(color=color_options.majority_color, showscale=False),
            )
        )

        pred_actual.update_layout(
            xaxis_title="Actual value",
            yaxis_title="Predicted value",
            xaxis=dict(showticklabels=True),
            yaxis=dict(showticklabels=True),
        )

        pred_actual_json = json.loads(pred_actual.to_json())

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1 if current_data is not None else 2,
            params={
                "data": pred_actual_json["data"],
                "layout": pred_actual_json["layout"],
            },
            additionalGraphs=[],
        )



import logging
import os
import platform

import requests

import evidently


class TelemetrySender:
    def __init__(self, address):
        self.address = address
        self.env = _collect_environment()
        self.evi = _collect_package()

    def send(self, usage):
        collected = dict(
            environment=self.env,
            evidently=self.evi,
            usage=usage,
        )
        try:
            requests.post(self.address, json=collected, timeout=3)
        except Exception as error:  # pylint: disable=broad-except
            logging.warning(f"failed to send telemetry: {error}")


def _collect_environment():
    return dict(
        python=dict(
            version=platform.python_version(),
            interpreter=platform.python_implementation(),
            conda="CONDA_DEFAULT_ENV" in os.environ,
            venv="VIRTUAL_ENV" in os.environ,
        ),
        os=platform.platform(),
    )


def _collect_package():
    return dict(version=evidently.__version__)



import dataclasses
from typing import List
from typing import Optional

import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import HistogramData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import get_histogram_figure_with_quantile
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import get_distribution_for_column


@dataclasses.dataclass
class ColumnQuantileMetricResult:
    column_name: str
    # range of the quantile (from 0 to 1)
    quantile: float
    # calculated value of the quantile in current data
    current: float
    # calculated value of the quantile in reference data
    # distribution for the column in current
    current_distribution: Distribution
    reference: Optional[float] = None
    # distribution for the column in reference
    reference_distribution: Optional[Distribution] = None


class ColumnQuantileMetric(Metric[ColumnQuantileMetricResult]):
    """Calculates quantile with specified range"""

    column_name: str
    quantile: float

    def __init__(self, column_name: str, quantile: float) -> None:
        self.quantile = quantile
        self.column_name = column_name

    def calculate(self, data: InputData) -> ColumnQuantileMetricResult:
        if not 0 < self.quantile <= 1:
            raise ValueError("Quantile should all be in the interval (0, 1].")

        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' is not in current data.")

        current_column = data.current_data[self.column_name]

        if not pd.api.types.is_numeric_dtype(current_column.dtype):
            raise ValueError(f"Column '{self.column_name}' in current data is not numeric.")

        current_quantile = data.current_data[self.column_name].quantile(self.quantile)

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column '{self.column_name}' is not in reference data.")

            reference_column = data.reference_data[self.column_name]

            if not pd.api.types.is_numeric_dtype(reference_column.dtype):
                raise ValueError(f"Column '{self.column_name}' in reference data is not numeric.")

            reference_quantile = reference_column.quantile(self.quantile)

        else:
            reference_column = None
            reference_quantile = None

        distributions = get_distribution_for_column(
            column_type="num", current=current_column, reference=reference_column
        )
        return ColumnQuantileMetricResult(
            column_name=self.column_name,
            current=current_quantile,
            quantile=self.quantile,
            current_distribution=distributions[0],
            reference_distribution=distributions[1],
            reference=reference_quantile,
        )


@default_renderer(wrap_type=ColumnQuantileMetric)
class ColumnQuantileMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnQuantileMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result.pop("current_distribution", None)
        result.pop("reference_distribution", None)
        return result

    @staticmethod
    def _get_counters(metric_result: ColumnQuantileMetricResult) -> BaseWidgetInfo:
        counters = [
            CounterData.float(label="Quantile", value=metric_result.quantile, precision=3),
            CounterData.float(label="Quantile value (current)", value=metric_result.current, precision=3),
        ]

        if metric_result.reference is not None:
            counters.append(
                CounterData.float(label="Quantile value (reference)", value=metric_result.reference, precision=3),
            )
        return counter(counters=counters)

    def _get_histogram(self, metric_result: ColumnQuantileMetricResult) -> BaseWidgetInfo:
        if metric_result.reference_distribution is not None:
            reference_histogram_data: Optional[HistogramData] = HistogramData(
                name="reference",
                x=list(metric_result.reference_distribution.x),
                y=list(metric_result.reference_distribution.y),
            )

        else:
            reference_histogram_data = None

        if metric_result.reference is not None:
            reference_quantile: Optional[float] = metric_result.reference

        else:
            reference_quantile = None

        figure = get_histogram_figure_with_quantile(
            current=HistogramData(
                name="current",
                x=list(metric_result.current_distribution.x),
                y=list(metric_result.current_distribution.y),
            ),
            reference=reference_histogram_data,
            current_quantile=metric_result.current,
            reference_quantile=reference_quantile,
            color_options=self.color_options,
        )
        figure.update_layout(
            yaxis_title="count",
            xaxis_title=metric_result.column_name,
        )
        return plotly_figure(title="", figure=figure)

    def render_html(self, obj: ColumnQuantileMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        column_name = metric_result.column_name
        return [
            header_text(label=f"Column '{column_name}'. Quantile."),
            self._get_counters(metric_result),
            self._get_histogram(metric_result),
        ]



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.num_target_drift_analyzer import NumTargetDriftAnalyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions
from evidently.options import QualityMetricsOptions


class NumOutputValuesWidget(Widget):
    def __init__(self, title: str, kind: str = "target"):
        super().__init__(title)
        self.title = title
        self.kind = kind  # target or prediction

    def analyzers(self):
        return [NumTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = NumTargetDriftAnalyzer.get_results(analyzers_results)
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)
        conf_interval_n_sigmas = quality_metrics_options.conf_interval_n_sigmas

        if current_data is None:
            raise ValueError("current_data should be present")

        if self.kind == "target":
            if results.columns.utility_columns.target is None:
                return None

            column_name = results.columns.utility_columns.target

        elif self.kind == "prediction":
            if results.columns.utility_columns.prediction is None:
                return None

            if not isinstance(results.columns.utility_columns.prediction, str):
                raise ValueError(
                    f"Widget [{self.title}] requires one str value for 'prediction' column"
                )

            column_name = results.columns.utility_columns.prediction

        else:
            raise ValueError(
                f"Widget [{self.title}] requires 'target' or 'prediction' kind parameter value"
            )

        utility_columns_date = results.columns.utility_columns.date
        # plot values
        reference_mean = np.mean(reference_data[column_name])
        reference_std = np.std(reference_data[column_name], ddof=1)
        x_title = "Timestamp" if utility_columns_date else "Index"

        output_values = go.Figure()

        output_values.add_trace(
            go.Scattergl(
                x=reference_data[utility_columns_date]
                if utility_columns_date
                else reference_data.index,
                y=reference_data[column_name],
                mode="markers",
                name="Reference",
                marker=dict(size=6, color=color_options.get_reference_data_color()),
            )
        )

        output_values.add_trace(
            go.Scattergl(
                x=current_data[utility_columns_date]
                if utility_columns_date
                else current_data.index,
                y=current_data[column_name],
                mode="markers",
                name="Current",
                marker=dict(size=6, color=color_options.get_current_data_color()),
            )
        )

        if utility_columns_date:
            x0 = current_data[utility_columns_date].sort_values()[1]
        else:
            x0 = current_data.index.sort_values()[1]

        output_values.add_trace(
            go.Scatter(
                x=[x0, x0],
                y=[
                    reference_mean - conf_interval_n_sigmas * reference_std,
                    reference_mean + conf_interval_n_sigmas * reference_std,
                ],
                mode="markers",
                name="Current",
                marker=dict(
                    size=0.01, color=color_options.non_visible_color, opacity=0.005
                ),
                showlegend=False,
            )
        )

        output_values.update_layout(
            xaxis_title=x_title,
            yaxis_title=self.kind.title() + " Value",
            showlegend=True,
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
            shapes=[
                dict(
                    type="rect",
                    # x-reference is assigned to the x-values
                    xref="paper",
                    # y-reference is assigned to the plot paper [0,1]
                    yref="y",
                    x0=0,
                    y0=reference_mean - conf_interval_n_sigmas * reference_std,
                    x1=1,
                    y1=reference_mean + conf_interval_n_sigmas * reference_std,
                    fillcolor=color_options.fill_color,
                    opacity=0.5,
                    layer="below",
                    line_width=0,
                ),
                dict(
                    type="line",
                    name="Reference",
                    xref="paper",
                    yref="y",
                    x0=0,  # min(testset_agg_by_date.index),
                    y0=reference_mean,
                    x1=1,  # max(testset_agg_by_date.index),
                    y1=reference_mean,
                    line=dict(color=color_options.zero_line_color, width=3),
                ),
            ],
        )

        output_values_json = json.loads(output_values.to_json())

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1,
            params={
                "data": output_values_json["data"],
                "layout": output_values_json["layout"],
            },
        )



from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.bar_widget import BarWidget
from evidently.dashboard.widgets.counter_widget import CounterWidget
from evidently.dashboard.widgets.expandable_list_widget import ExpandableListWidget
from evidently.dashboard.widgets.percent_widget import PercentWidget
from evidently.dashboard.widgets.test_suite_widget import TestSuiteWidget
from evidently.dashboard.widgets.text_widget import TextWidget


class WidgetGalleryTab(Tab):
    widgets = [
        (BarWidget(""), Verbose.ALWAYS),
        (CounterWidget(""), Verbose.ALWAYS),
        (PercentWidget(""), Verbose.ALWAYS),
        (ExpandableListWidget("Some title"), Verbose.ALWAYS),
        (TextWidget("Some title"), Verbose.ALWAYS),
        (TestSuiteWidget(""), Verbose.ALWAYS),
    ]



import dataclasses
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import WidgetSize
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.utils.data_operations import process_columns
from evidently.utils.data_operations import recognize_column_type
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import get_distribution_for_column


@dataclasses.dataclass
class ColumnDistributionMetricResult:
    column_name: str
    current: Distribution
    reference: Optional[Distribution] = None


class ColumnDistributionMetric(Metric[ColumnDistributionMetricResult]):
    """Calculates distribution for the column"""

    column_name: str

    def __init__(
        self,
        column_name: str,
    ) -> None:
        self.column_name = column_name

    def calculate(self, data: InputData) -> ColumnDistributionMetricResult:
        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' was not found in current data.")

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column '{self.column_name}' was not found in reference data.")

        columns = process_columns(data.current_data, data.column_mapping)
        column_type = recognize_column_type(dataset=data.current_data, column_name=self.column_name, columns=columns)
        current_column = data.current_data[self.column_name]
        reference_column = None
        if data.reference_data is not None:
            reference_column = data.reference_data[self.column_name]
        current, reference = get_distribution_for_column(
            column_type=column_type,
            current=current_column,
            reference=reference_column,
        )

        return ColumnDistributionMetricResult(
            column_name=self.column_name,
            current=current,
            reference=reference,
        )


@default_renderer(wrap_type=ColumnDistributionMetric)
class ColumnDistributionMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnDistributionMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result.pop("current", None)
        result.pop("reference", None)
        return result

    def render_html(self, obj: ColumnDistributionMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        distr_fig = get_distribution_plot_figure(
            current_distribution=metric_result.current,
            reference_distribution=metric_result.reference,
            color_options=self.color_options,
        )

        result = [
            header_text(label=f"Distribution for column '{metric_result.column_name}'."),
            plotly_figure(title="", figure=distr_fig, size=WidgetSize.FULL),
        ]
        return result



from typing import Tuple

import pandas as pd

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.utils.data_drift_utils import calculate_text_drift_score


def _text_content_drift(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    return calculate_text_drift_score(reference_data, current_data, p_value=threshold)


text_content_drift_stat_test = StatTest(
    name="text_content_drift",
    display_name="Text content drift",
    func=_text_content_drift,
    allowed_feature_types=["text"],
    default_threshold=0.05,
)

register_stattest(text_content_drift_stat_test)



import dataclasses
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd

from evidently import ColumnMapping
from evidently import TaskType


@dataclasses.dataclass
class _InputData:
    reference: Optional[pd.DataFrame]
    current: pd.DataFrame


class ColumnType(Enum):
    Categorical = "cat"
    Numerical = "num"
    Datetime = "datetime"
    Text = "text"


@dataclasses.dataclass
class ColumnDefinition:
    column_name: str
    column_type: ColumnType


@dataclasses.dataclass
class PredictionColumns:
    predicted_values: Optional[ColumnDefinition] = None
    prediction_probas: Optional[List[ColumnDefinition]] = None

    def get_columns_list(self) -> List[ColumnDefinition]:
        result = [self.predicted_values]
        if self.prediction_probas is not None:
            result.extend(self.prediction_probas)
        return [col for col in result if col is not None]


def _check_filter(column: ColumnDefinition, utility_columns: List[str], filter_def: str) -> bool:
    if filter_def == "all":
        return True
    if filter_def == "categorical_columns":
        return column.column_type == ColumnType.Categorical
    if filter_def == "numerical_columns":
        return column.column_type == ColumnType.Numerical
    if filter_def == "datetime_columns":
        return column.column_type == ColumnType.Datetime
    if filter_def == "text_columns":
        return column.column_type == ColumnType.Text
    if filter_def == "all_features":
        return column.column_name not in utility_columns
    if filter_def == "text_features":
        return column.column_type == ColumnType.Text and column.column_name not in utility_columns
    if filter_def == "categorical_features":
        return column.column_type == ColumnType.Categorical and column.column_name not in utility_columns
    if filter_def == "numerical_features":
        return column.column_type == ColumnType.Numerical and column.column_name not in utility_columns
    if filter_def == "datetime_features":
        return column.column_type == ColumnType.Datetime and column.column_name not in utility_columns
    raise ValueError(f"Unknown filter: {filter_def}")


@dataclasses.dataclass
class DataDefinition:
    _columns: Dict[str, ColumnDefinition]
    _target: Optional[ColumnDefinition]
    _prediction_columns: Optional[PredictionColumns]
    _id_column: Optional[ColumnDefinition]
    _datetime_column: Optional[ColumnDefinition]

    _task: Optional[str]
    _classification_labels: Optional[Sequence[str]]

    def __init__(
        self,
        columns: List[ColumnDefinition],
        target: Optional[ColumnDefinition],
        prediction_columns: Optional[PredictionColumns],
        id_column: Optional[ColumnDefinition],
        datetime_column: Optional[ColumnDefinition],
        task: Optional[str],
        classification_labels: Optional[Sequence[str]],
    ):
        self._columns = {column.column_name: column for column in columns}
        self._id_column = id_column
        self._datetime_column = datetime_column
        self._task = task
        self._target = target
        self._prediction_columns = prediction_columns
        self._classification_labels = classification_labels

    def get_column(self, column_name: str) -> ColumnDefinition:
        return self._columns[column_name]

    def get_columns(self, filter_def: str = "all") -> List[ColumnDefinition]:
        if self._prediction_columns is not None:
            prediction = self._prediction_columns.get_columns_list()
        else:
            prediction = []
        utility_columns = [
            col.column_name
            for col in [
                self._id_column,
                self._datetime_column,
                self._target,
                *prediction,
            ]
            if col is not None
        ]
        return [column for column in self._columns.values() if _check_filter(column, utility_columns, filter_def)]

    def get_target_column(self) -> Optional[ColumnDefinition]:
        return self._target

    def get_prediction_columns(self) -> Optional[PredictionColumns]:
        return self._prediction_columns

    def get_id_column(self) -> Optional[ColumnDefinition]:
        return self._id_column

    def get_datetime_column(self) -> Optional[ColumnDefinition]:
        return self._datetime_column

    def task(self) -> Optional[str]:
        return self._task

    def classification_labels(self) -> Optional[Sequence[str]]:
        return self._classification_labels


def _process_column(
    column_name: Optional[str],
    data: _InputData,
    if_partially_present: str = "raise",
    predefined_type: Optional[ColumnType] = None,
) -> Optional[ColumnDefinition]:
    if column_name is None:
        return None
    presence = _get_column_presence(column_name, data)
    if presence == ColumnPresenceState.Missing:
        return None
    if presence == ColumnPresenceState.Partially:
        if if_partially_present == "raise":
            raise ValueError(f"Column ({column_name}) is partially present in data")
        if if_partially_present == "skip":
            return None
        if if_partially_present == "keep":
            pass
    column_type = predefined_type if predefined_type is not None else _get_column_type(column_name, data)
    return ColumnDefinition(column_name, column_type)


def _prediction_column(
    prediction: Optional[Union[str, int, Sequence[int], Sequence[str]]],
    target_type: Optional[ColumnType],
    target_names: Optional[List[str]],
    task: Optional[str],
    data: _InputData,
) -> Optional[PredictionColumns]:
    if prediction is None:
        return None
    if isinstance(prediction, str):
        prediction_present = _get_column_presence(prediction, data)
        if prediction_present == ColumnPresenceState.Missing:
            return None
        if prediction_present == ColumnPresenceState.Partially:
            raise ValueError(f"Prediction column ({prediction}) is partially present in data")
        prediction_type = _get_column_type(prediction, data)
        if task == TaskType.CLASSIFICATION_TASK:
            if prediction_type == ColumnType.Categorical:
                return PredictionColumns(predicted_values=ColumnDefinition(prediction, prediction_type))
            if prediction_type == ColumnType.Numerical:
                return PredictionColumns(prediction_probas=[ColumnDefinition(prediction, prediction_type)])
            raise ValueError(f"Unexpected type for prediction column ({prediction}) (it is {prediction_type})")
        if task == TaskType.REGRESSION_TASK:
            if prediction_type == ColumnType.Categorical:
                raise ValueError("Prediction type is categorical but task is regression")
            if prediction_type == ColumnType.Numerical:
                return PredictionColumns(predicted_values=ColumnDefinition(prediction, prediction_type))
        if task is None:
            if prediction_type == ColumnType.Numerical and target_type == ColumnType.Categorical:
                # probably this is binary with single column of probabilities
                return PredictionColumns(prediction_probas=[ColumnDefinition(prediction, prediction_type)])
            return PredictionColumns(predicted_values=ColumnDefinition(prediction, prediction_type))
    if isinstance(prediction, list):
        if target_names is not None:
            if prediction != target_names:
                raise ValueError("List of prediction columns should be equal to target_names if both set")
        presence = [_get_column_presence(column, data) for column in prediction]
        if all([item == ColumnPresenceState.Missing for item in presence]):
            return None
        if all([item == ColumnPresenceState.Present for item in presence]):
            prediction_defs = [ColumnDefinition(column, _get_column_type(column, data)) for column in prediction]
            if any([item.column_type != ColumnType.Numerical for item in prediction_defs]):
                raise ValueError(f"Some prediction columns have incorrect types {prediction_defs}")
            return PredictionColumns(prediction_probas=prediction_defs)
    raise ValueError("Unexpected type for prediction field in column_mapping")


def _filter_by_type(column: Optional[ColumnDefinition], column_type: ColumnType, exclude: List[str]) -> bool:
    return column is not None and column.column_type == column_type and column.column_name not in exclude


def create_data_definition(
    reference_data: Optional[pd.DataFrame],
    current_data: pd.DataFrame,
    mapping: ColumnMapping,
) -> DataDefinition:
    data = _InputData(reference_data, current_data)
    id_column = _process_column(mapping.id, data)
    target_column = _process_column(mapping.target, data)
    datetime_column = _process_column(mapping.datetime, data)

    prediction_columns = _prediction_column(
        mapping.prediction,
        target_column.column_type if target_column is not None else None,
        mapping.target_names,
        mapping.task,
        data,
    )

    prediction_cols = prediction_columns.get_columns_list() if prediction_columns is not None else []

    all_columns = [
        id_column,
        datetime_column,
        target_column,
        *prediction_cols,
    ]
    utility_column_names = [column.column_name for column in all_columns if column is not None]
    data_columns = set(data.current.columns) | (set(data.reference.columns) if data.reference is not None else set())
    col_defs = [_process_column(column_name, data, if_partially_present="skip") for column_name in data_columns]

    if mapping.numerical_features is None:
        num = [column for column in col_defs if _filter_by_type(column, ColumnType.Numerical, utility_column_names)]
        all_columns.extend(num)
    else:
        all_columns.extend(
            [
                _process_column(column_name, data, predefined_type=ColumnType.Numerical)
                for column_name in mapping.numerical_features
                if column_name not in utility_column_names
            ]
        )

    if mapping.categorical_features is None:
        cat = [column for column in col_defs if _filter_by_type(column, ColumnType.Categorical, utility_column_names)]
        all_columns.extend(cat)
    else:
        all_columns.extend(
            [
                _process_column(column_name, data, predefined_type=ColumnType.Categorical)
                for column_name in mapping.categorical_features
                if column_name not in utility_column_names
            ]
        )

    if mapping.datetime_features is None:
        dt = [column for column in col_defs if _filter_by_type(column, ColumnType.Datetime, utility_column_names)]
        all_columns.extend(dt)
    else:
        all_columns.extend(
            [
                _process_column(column_name, data, predefined_type=ColumnType.Datetime)
                for column_name in mapping.datetime_features
                if column_name not in utility_column_names
            ]
        )

    if mapping.text_features is not None:
        all_columns.extend(
            [
                _process_column(column_name, data, predefined_type=ColumnType.Text)
                for column_name in mapping.text_features
                if column_name not in utility_column_names
            ]
        )
    task = mapping.task
    if task is None:
        if target_column is None:
            task = None
        elif target_column.column_type == ColumnType.Categorical:
            task = TaskType.CLASSIFICATION_TASK
        elif target_column.column_type == ColumnType.Numerical:
            task = TaskType.REGRESSION_TASK
        else:
            task = None

    return DataDefinition(
        columns=[col for col in all_columns if col is not None],
        id_column=id_column,
        datetime_column=datetime_column,
        target=target_column,
        prediction_columns=prediction_columns,
        task=task,
        classification_labels=mapping.target_names,
    )


class ColumnPresenceState(Enum):
    Present = 0
    Partially = 1
    Missing = 2


def _get_column_presence(column_name: str, data: _InputData) -> ColumnPresenceState:
    if column_name in data.current.columns:
        if data.reference is None or column_name in data.reference.columns:
            return ColumnPresenceState.Present
        return ColumnPresenceState.Partially
    if data.reference is None or column_name not in data.reference.columns:
        return ColumnPresenceState.Missing
    return ColumnPresenceState.Partially


NUMBER_UNIQUE_AS_CATEGORICAL = 5


def _get_column_type(column_name: str, data: _InputData) -> ColumnType:
    ref_type = None
    ref_unique = None
    if data.reference is not None and column_name in data.reference.columns:
        ref_type = data.reference[column_name].dtype
        ref_unique = data.reference[column_name].nunique()
    cur_type = None
    cur_unique = None
    if column_name in data.current.columns:
        cur_type = data.current[column_name].dtype
        cur_unique = data.current[column_name].nunique()
    if (
        ref_type is not None
        and cur_type is not None
        and (ref_type != cur_type and not np.can_cast(cur_type, ref_type) and not np.can_cast(ref_type, cur_type))
    ):
        raise ValueError(f"Column {column_name} have different types in reference {ref_type} and current {cur_type}")
    if pd.api.types.is_integer_dtype(cur_type if cur_type is not None else ref_type):
        nunique = ref_unique or cur_unique
        if nunique is not None and nunique <= NUMBER_UNIQUE_AS_CATEGORICAL:
            return ColumnType.Categorical
        return ColumnType.Numerical
    if pd.api.types.is_numeric_dtype(cur_type if cur_type is not None else ref_type):
        return ColumnType.Numerical
    if pd.api.types.is_datetime64_dtype(cur_type if cur_type is not None else ref_type):
        return ColumnType.Datetime
    return ColumnType.Categorical



import abc
import dataclasses
from abc import ABC
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from evidently.utils.generators import BaseGenerator
from evidently.utils.generators import make_generator_by_columns
from evidently.utils.types import ApproxValue
from evidently.utils.types import Numeric


@dataclasses.dataclass
class GroupData:
    id: str
    title: str
    description: str
    sort_index: int = 0
    severity: Optional[str] = None


@dataclasses.dataclass
class GroupTypeData:
    id: str
    title: str
    # possible values with description, if empty will use simple view (no severity, description and sorting).
    values: List[GroupData] = dataclasses.field(default_factory=list)

    def add_value(self, data: GroupData):
        self.values.append(data)


class GroupingTypes:
    ByFeature = GroupTypeData(
        "by_feature",
        "By feature",
        [
            GroupData(
                "no group",
                "Dataset-level tests",
                "Some tests cannot be grouped by feature",
            )
        ],
    )
    ByClass = GroupTypeData("by_class", "By class", [])
    TestGroup = GroupTypeData(
        "test_group",
        "By test group",
        [
            GroupData(
                "no group",
                "Ungrouped",
                "Some tests don’t belong to any group under the selected condition",
            )
        ],
    )
    TestType = GroupTypeData("test_type", "By test type", [])


DEFAULT_GROUP = [GroupingTypes.ByFeature, GroupingTypes.TestGroup, GroupingTypes.TestType, GroupingTypes.ByClass]


@dataclass
class TestResult:
    # Constants for test result status
    SUCCESS = "SUCCESS"  # the test was passed
    FAIL = "FAIL"  # success pass for the test
    WARNING = "WARNING"  # the test was passed, but we have some issues during the execution
    ERROR = "ERROR"  # cannot calculate the test result, no data
    SKIPPED = "SKIPPED"  # the test was skipped

    # short name/title from the test class
    name: str
    # what was checked, what threshold (current value 13 is not ok with condition less than 5)
    description: str
    # status of the test result
    status: str
    # grouping parameters
    groups: Dict[str, str] = dataclasses.field(default_factory=dict)

    def set_status(self, status: str, description: Optional[str] = None) -> None:
        self.status = status

        if description is not None:
            self.description = description

    def mark_as_fail(self, description: Optional[str] = None):
        self.set_status(self.FAIL, description=description)

    def mark_as_error(self, description: Optional[str] = None):
        self.set_status(self.ERROR, description=description)

    def mark_as_success(self, description: Optional[str] = None):
        self.set_status(self.SUCCESS, description=description)

    def mark_as_warning(self, description: Optional[str] = None):
        self.set_status(self.WARNING, description=description)

    def is_passed(self):
        return self.status in [self.SUCCESS, self.WARNING]


class Test:
    """
    all fields in test class with type that is subclass of Metric would be used as dependencies of test.
    """

    name: str
    group: str
    context = None

    @abc.abstractmethod
    def check(self) -> TestResult:
        raise NotImplementedError()

    def set_context(self, context):
        self.context = context

    def get_result(self):
        if self.context is None:
            raise ValueError("No context is set")
        result = self.context.test_results.get(self, None)
        if result is None:
            raise ValueError(f"No result found for metric {self} of type {type(self).__name__}")
        return result


@dataclass
class TestValueCondition:
    """
    Class for processing a value conditions - should it be less, greater than, equals and so on.

    An object of the class stores specified conditions and can be used for checking a value by them.
    """

    eq: Optional[Numeric] = None
    gt: Optional[Numeric] = None
    gte: Optional[Numeric] = None
    is_in: Optional[List[Union[Numeric, str, bool]]] = None
    lt: Optional[Numeric] = None
    lte: Optional[Numeric] = None
    not_eq: Optional[Numeric] = None
    not_in: Optional[List[Union[Numeric, str, bool]]] = None

    def has_condition(self) -> bool:
        """
        Checks if we have a condition in the object and returns True in this case.

        If we have no conditions - returns False.
        """
        return any(
            value is not None
            for value in (self.eq, self.gt, self.gte, self.is_in, self.lt, self.lte, self.not_in, self.not_eq)
        )

    def check_value(self, value: Numeric) -> bool:
        result = True

        if self.eq is not None and result:
            result = value == self.eq

        if self.gt is not None and result:
            result = value > self.gt

        if self.gte is not None and result:
            result = value >= self.gte

        if self.is_in is not None and result:
            result = value in self.is_in

        if self.lt is not None and result:
            result = value < self.lt

        if self.lte is not None and result:
            result = value <= self.lte

        if self.not_eq is not None and result:
            result = value != self.not_eq

        if self.not_in is not None and result:
            result = value not in self.not_in

        return result

    def __str__(self) -> str:
        conditions = []
        operations = ["eq", "gt", "gte", "lt", "lte", "not_eq", "is_in", "not_in"]

        for op in operations:
            value = getattr(self, op)

            if value is None:
                continue

            if isinstance(value, (float, ApproxValue)):
                conditions.append(f"{op}={value:.3g}")

            else:
                conditions.append(f"{op}={value}")

        return f"{' and '.join(conditions)}"

    def as_dict(self) -> dict:
        result = {}
        operations = ["eq", "gt", "gte", "lt", "lte", "not_eq", "is_in", "not_in"]

        for op in operations:
            value = getattr(self, op)

            if value is not None:
                result[op] = value

        return result


class BaseConditionsTest(Test, ABC):
    """
    Base class for all tests with a condition
    """

    condition: TestValueCondition

    def __init__(
        self,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        self.condition = TestValueCondition(
            eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in
        )


class BaseCheckValueTest(BaseConditionsTest):
    """
    Base class for all tests with checking a value condition
    """

    value: Numeric

    @abc.abstractmethod
    def calculate_value_for_test(self) -> Optional[Any]:
        """Method for getting the checking value.

        Define it in a child class"""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_description(self, value: Numeric) -> str:
        """Method for getting a description that we can use.
        The description can use the checked value.

        Define it in a child class"""
        raise NotImplementedError()

    def get_condition(self) -> TestValueCondition:
        return self.condition

    def groups(self) -> Dict[str, str]:
        return {}

    def check(self):
        result = TestResult(name=self.name, description="The test was not launched", status=TestResult.SKIPPED)
        value = self.calculate_value_for_test()
        self.value = value
        result.description = self.get_description(value)

        try:
            if value is None:
                result.mark_as_error()

            else:
                condition = self.get_condition()

                if condition is None:
                    raise ValueError

                condition_check_result = condition.check_value(value)

                if condition_check_result:
                    result.mark_as_success()

                else:
                    result.mark_as_fail()

        except ValueError:
            result.mark_as_error("Cannot calculate the condition")

        result.groups.update(self.groups())
        return result


def generate_column_tests(
    test_class: Type[Test],
    columns: Optional[Union[str, list]] = None,
    parameters: Optional[Dict] = None,
) -> BaseGenerator:
    """Function for generating tests for columns"""
    return make_generator_by_columns(
        base_class=test_class,
        columns=columns,
        parameters=parameters,
    )



#!/usr/bin/env python
# coding: utf-8
from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.cat_output_drift_widget import CatOutputDriftWidget
from evidently.dashboard.widgets.cat_target_pred_feature_table_widget import (
    CatTargetPredFeatureTable,
)
from evidently.dashboard.widgets.prob_class_pred_distr_widget import (
    ProbClassPredDistrWidget,
)


class CatTargetDriftTab(Tab):
    widgets = [
        (CatOutputDriftWidget("Target Drift"), Verbose.ALWAYS),
        (CatOutputDriftWidget("Prediction Drift", "prediction"), Verbose.ALWAYS),
        (ProbClassPredDistrWidget("Reference: Probability Distribution"), Verbose.FULL),
        (
            ProbClassPredDistrWidget("Current: Probability Distribution", "current"),
            Verbose.FULL,
        ),
        (
            CatTargetPredFeatureTable("Target (Prediction) Behavior By Feature"),
            Verbose.FULL,
        ),
    ]



import dataclasses
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_drift import ColumnDataDriftMetrics
from evidently.calculations.data_drift import get_dataset_drift
from evidently.calculations.data_drift import get_one_column_drift
from evidently.calculations.stattests import PossibleStatTestType
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import BaseWidgetInfo
from evidently.options import DataDriftOptions
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import ColumnDefinition
from evidently.renderers.html_widgets import ColumnType
from evidently.renderers.html_widgets import RichTableDataRow
from evidently.renderers.html_widgets import RowDetails
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import rich_table_data
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import ColumnType as ColumnType_data
from evidently.utils.data_preprocessing import DataDefinition
from evidently.utils.visualizations import plot_scatter_for_data_drift


@dataclass
class TextDescriptorsDriftMetricResults:
    number_of_columns: int
    number_of_drifted_columns: int
    share_of_drifted_columns: float
    dataset_drift: bool
    drift_by_columns: Dict[str, ColumnDataDriftMetrics]
    dataset_columns: DatasetColumns


class TextDescriptorsDriftMetric(Metric[TextDescriptorsDriftMetricResults]):
    column_name: str
    options: DataDriftOptions
    generated_text_features: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]

    def __init__(
        self,
        column_name: str,
        stattest: Optional[PossibleStatTestType] = None,
        stattest_threshold: Optional[float] = None,
    ):
        self.column_name = column_name
        self.options = DataDriftOptions(all_features_stattest=stattest, all_features_threshold=stattest_threshold)
        self.generated_text_features = {}

    def required_features(self, data_definition: DataDefinition):
        column_type = data_definition.get_column(self.column_name).column_type
        if column_type == ColumnType_data.Text:
            self.generated_text_features = {}
            self.generated_text_features["Text Length"] = TextLength(self.column_name)
            self.generated_text_features["Non Letter Character %"] = NonLetterCharacterPercentage(self.column_name)
            self.generated_text_features["OOV %"] = OOVWordsPercentage(self.column_name)
            return list(self.generated_text_features.values())
        return []

    def get_parameters(self) -> tuple:
        return self.column_name, self.options

    def calculate(self, data: InputData) -> TextDescriptorsDriftMetricResults:
        if data.reference_data is None:
            raise ValueError("Reference dataset should be present")
        curr_text_df = pd.concat(
            [data.get_current_column(x.feature_name()) for x in list(self.generated_text_features.values())],
            axis=1,
        )
        curr_text_df.columns = list(self.generated_text_features.keys())

        ref_text_df = pd.concat(
            [data.get_reference_column(x.feature_name()) for x in list(self.generated_text_features.values())],
            axis=1,
        )
        ref_text_df.columns = list(self.generated_text_features.keys())
        # text_dataset_columns = DatasetColumns(num_feature_names=curr_text_df.columns)
        text_dataset_columns = process_columns(ref_text_df, ColumnMapping(numerical_features=ref_text_df.columns))

        drift_by_columns = {}
        for col in curr_text_df.columns:
            drift_by_columns[col] = get_one_column_drift(
                current_data=curr_text_df,
                reference_data=ref_text_df,
                column_name=col,
                options=self.options,
                dataset_columns=text_dataset_columns,
            )
        dataset_drift = get_dataset_drift(drift_by_columns, 0)

        return TextDescriptorsDriftMetricResults(
            number_of_columns=curr_text_df.shape[1],
            number_of_drifted_columns=dataset_drift.number_of_drifted_columns,
            share_of_drifted_columns=dataset_drift.dataset_drift_score,
            dataset_drift=dataset_drift.dataset_drift,
            drift_by_columns=drift_by_columns,
            dataset_columns=text_dataset_columns,
        )


@default_renderer(wrap_type=TextDescriptorsDriftMetric)
class DataDriftTableRenderer(MetricRenderer):
    def render_json(self, obj: TextDescriptorsDriftMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())

        # remove pandas dataset values and other useless for JSON fields
        result.pop("dataset_columns", None)

        for column_name, data in result["drift_by_columns"].items():
            data.pop("current_distribution", None)
            data.pop("reference_distribution", None)
            data.pop("current_small_distribution", None)
            data.pop("reference_small_distribution", None)
            data.pop("current_correlations", None)
            data.pop("reference_correlations", None)
            data.pop("current_scatter", None)
            data.pop("x_name", None)
            data.pop("plot_shape", None)

        return result

    def _generate_column_params(self, column_name: str, data: ColumnDataDriftMetrics) -> Optional[RichTableDataRow]:
        details = RowDetails()
        if (
            data.current_small_distribution is None
            or data.reference_small_distribution is None
            or data.current_distribution is None
        ):
            return None

        current_small_hist = data.current_small_distribution
        ref_small_hist = data.reference_small_distribution
        data_drift = "Detected" if data.drift_detected else "Not Detected"
        if (
            data.column_type == "num"
            and data.current_scatter is not None
            and data.x_name is not None
            and data.plot_shape is not None
        ):
            scatter_fig = plot_scatter_for_data_drift(
                curr_y=data.current_scatter[data.column_name],
                curr_x=data.current_scatter[data.x_name],
                y0=data.plot_shape["y0"],
                y1=data.plot_shape["y1"],
                y_name=data.column_name,
                x_name=data.x_name,
                color_options=self.color_options,
            )
            scatter = plotly_figure(title="", figure=scatter_fig)
            details.with_part("DATA DRIFT", info=scatter)
            fig = get_distribution_plot_figure(
                current_distribution=data.current_distribution,
                reference_distribution=data.reference_distribution,
                color_options=self.color_options,
            )
            distribution = plotly_figure(title="", figure=fig)
            details.with_part("DATA DISTRIBUTION", info=distribution)
            return RichTableDataRow(
                details=details,
                fields={
                    "column_name": column_name,
                    "column_type": data.column_type,
                    "stattest_name": data.stattest_name,
                    "reference_distribution": {"x": list(ref_small_hist[1]), "y": list(ref_small_hist[0])},
                    "current_distribution": {"x": list(current_small_hist[1]), "y": list(current_small_hist[0])},
                    "data_drift": data_drift,
                    "drift_score": round(data.drift_score, 6),
                },
            )
        else:
            return None

    def render_html(self, obj: TextDescriptorsDriftMetric) -> List[BaseWidgetInfo]:
        results = obj.get_result()
        color_options = self.color_options

        # set params data
        params_data = []

        # sort columns by drift score
        columns = sorted(
            results.drift_by_columns.keys(),
            key=lambda x: results.drift_by_columns[x].drift_score,
            reverse=True,
        )

        for column_name in columns:
            column_params = self._generate_column_params(column_name, results.drift_by_columns[column_name])

            if column_params is not None:
                params_data.append(column_params)

        drift_percents = round(results.share_of_drifted_columns * 100, 3)

        return [
            header_text(label=f"Text Descriptors Drift for column '{obj.column_name}'"),
            rich_table_data(
                title=f"Drift is detected for {drift_percents}% of columns "
                f"({results.number_of_drifted_columns} out of {results.number_of_columns}).",
                columns=[
                    ColumnDefinition("Column", "column_name"),
                    ColumnDefinition("Type", "column_type"),
                    ColumnDefinition(
                        "Reference Distribution",
                        "reference_distribution",
                        ColumnType.HISTOGRAM,
                        options={"xField": "x", "yField": "y", "color": color_options.primary_color},
                    ),
                    ColumnDefinition(
                        "Current Distribution",
                        "current_distribution",
                        ColumnType.HISTOGRAM,
                        options={"xField": "x", "yField": "y", "color": color_options.primary_color},
                    ),
                    ColumnDefinition("Data Drift", "data_drift"),
                    ColumnDefinition("Stat Test", "stattest_name"),
                    ColumnDefinition("Drift Score", "drift_score"),
                ],
                data=params_data,
            ),
        ]



from typing import Optional

from evidently.base_metric import InputData
from evidently.calculations.stattests import PossibleStatTestType
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAccuracyScore
from evidently.tests import TestColumnDrift
from evidently.tests import TestF1Score
from evidently.tests import TestLogLoss
from evidently.tests import TestNumberOfRows
from evidently.tests import TestPrecisionByClass
from evidently.tests import TestRecallByClass
from evidently.tests import TestRocAuc
from evidently.utils.data_operations import DatasetColumns


class MulticlassClassificationTestPreset(TestPreset):
    """
    Multiclass Classification tests.

    Args:
        stattest: statistical test for target drift test.
        stattest_threshold: threshold for statistical test for target drift test.

    Contains tests:
    - `TestAccuracyScore`
    - `TestF1Score`
    - `TestPrecisionByClass` for each class in data
    - `TestRecallByClass` for each class in data
    - `TestNumberOfRows`
    - `TestColumnValueDrift`
    - `TestRocAuc`
    - `TestLogLoss`
    """

    stattest: Optional[PossibleStatTestType]
    stattest_threshold: Optional[float]

    def __init__(
        self,
        stattest: Optional[PossibleStatTestType] = None,
        stattest_threshold: Optional[float] = None,
    ):
        super().__init__()

        self.stattest = stattest
        self.stattest_threshold = stattest_threshold

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        target = columns.utility_columns.target

        if target is None:
            raise ValueError("Target column should be set in mapping and be present in data")

        labels = set(data.current_data[target].unique())

        if data.reference_data is not None:
            labels = labels | set(data.reference_data[target].unique())

        tests = [
            TestAccuracyScore(),
            TestF1Score(),
            *[TestPrecisionByClass(str(label)) for label in labels],
            *[TestRecallByClass(str(label)) for label in labels],
            TestNumberOfRows(),
            TestColumnDrift(column_name=target, stattest=self.stattest, stattest_threshold=self.stattest_threshold),
        ]

        prediction_columns = data.data_definition.get_prediction_columns()
        if prediction_columns is None or prediction_columns.prediction_probas is None:
            return tests
        else:
            return tests + [TestRocAuc(), TestLogLoss()]



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.model.widget import TabInfo


class ProbClassPRTableWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        results = ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
        utility_columns = results.columns.utility_columns

        if utility_columns.target is None or utility_columns.prediction is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )

            return None

        if self.dataset == "reference":
            metrics = results.reference_metrics

            if metrics is None:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {ProbClassificationPerformanceAnalyzer.__name__} but no data found"
                )

        elif self.dataset == "current":
            metrics = results.current_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] required 'current' or 'reference' dataset value"
            )

        if metrics is None:
            return None

        if len(utility_columns.prediction) <= 2:
            if metrics.pr_table is None:
                raise ValueError(f"Widget [{self.title}] got no pr_table value")

            if not isinstance(metrics.pr_table, list):
                raise ValueError(
                    f"Widget [{self.title}] got incorrect type for pr_table value"
                )

            pr_table_data: list = metrics.pr_table
            params_data = []

            for line in pr_table_data:
                count = line[1]
                prob = round(line[2], 2)
                top = round(line[0], 1)
                tp = line[3]
                fp = line[4]
                precision = round(line[5], 1)
                recall = round(line[6], 1)

                params_data.append(
                    {
                        "f1": float(top),
                        "f2": int(count),
                        "f3": float(prob),
                        "f4": int(tp),
                        "f5": int(fp),
                        "f6": float(precision),
                        "f7": float(recall),
                    }
                )

            widget_info = BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=1 if current_data is not None else 2,
                params={
                    "rowsPerPage": 21,
                    "columns": [
                        {"title": "Top(%)", "field": "f1", "sort": "asc"},
                        {"title": "Count", "field": "f2"},
                        {
                            "title": "Prob",
                            "field": "f3",
                        },
                        {"title": "TP", "field": "f4"},
                        {"title": "FP", "field": "f5"},
                        {"title": "Precision", "field": "f6"},
                        {"title": "Recall", "field": "f7"},
                    ],
                    "data": params_data,
                },
            )

        else:
            # create tables
            tabs = []

            for label in utility_columns.prediction:
                params_data = []

                if metrics.pr_table is None:
                    raise ValueError(f"Widget [{self.title}] got no pr_table value")

                if not isinstance(metrics.pr_table, dict):
                    raise ValueError(
                        f"Widget [{self.title}] got incorrect type of pr_table value"
                    )

                pr_table_data_list = metrics.pr_table[label]

                for line in pr_table_data_list:
                    count = line[1]
                    prob = round(line[2], 2)
                    top = round(line[0], 1)
                    tp = line[3]
                    fp = line[4]
                    precision = round(line[5], 1)
                    recall = round(line[6], 1)

                    params_data.append(
                        {
                            "f1": float(top),
                            "f2": int(count),
                            "f3": float(prob),
                            "f4": int(tp),
                            "f5": int(fp),
                            "f6": float(precision),
                            "f7": float(recall),
                        }
                    )

                tabs.append(
                    TabInfo(
                        id=label,
                        title=label,
                        widget=BaseWidgetInfo(
                            title="",
                            type="big_table",
                            size=2,  # if current_data is not None else 2,
                            params={
                                "rowsPerPage": 21,
                                "columns": [
                                    {"title": "Top(%)", "field": "f1", "sort": "asc"},
                                    {"title": "Count", "field": "f2"},
                                    {"title": "Prob", "field": "f3"},
                                    {"title": "TP", "field": "f4"},
                                    {"title": "FP", "field": "f5"},
                                    {"title": "Precision", "field": "f6"},
                                    {"title": "Recall", "field": "f7"},
                                ],
                                "data": params_data,
                            },
                            additionalGraphs=[],
                        ),
                    )
                )

            widget_info = BaseWidgetInfo(
                type="tabs",
                title=self.title,
                size=1 if current_data is not None else 2,
                tabs=tabs,
            )
        return widget_info



from abc import ABC
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from evidently import ColumnMapping
from evidently.base_metric import Metric
from evidently.base_metric import TResult
from evidently.calculations.classification_performance import PredictionData
from evidently.calculations.classification_performance import get_prediction_data
from evidently.calculations.classification_performance import k_probability_threshold
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns


def _cleanup_data(data: pd.DataFrame, dataset_columns: DatasetColumns) -> pd.DataFrame:
    target = dataset_columns.utility_columns.target
    prediction = dataset_columns.utility_columns.prediction
    subset = []
    if target is not None:
        subset.append(target)
    if prediction is not None and isinstance(prediction, list):
        subset += prediction
    if prediction is not None and isinstance(prediction, str):
        subset.append(prediction)
    if len(subset) > 0:
        return data.replace([np.inf, -np.inf], np.nan).dropna(axis=0, how="any", subset=subset)
    return data


class ThresholdClassificationMetric(Metric[TResult], ABC):
    probas_threshold: Optional[float]
    k: Optional[Union[float, int]]

    def __init__(self, probas_threshold: Optional[float], k: Optional[Union[float, int]]):
        if probas_threshold is not None and k is not None:
            raise ValueError(
                f"{self.__class__.__name__}: should provide only stattest_threshold or top_k argument, not both."
            )

        self.probas_threshold = probas_threshold
        self.k = k

    def get_target_prediction_data(
        self,
        data: pd.DataFrame,
        column_mapping: ColumnMapping,
    ) -> Tuple[pd.Series, PredictionData]:
        dataset_columns = process_columns(data, column_mapping)
        data = _cleanup_data(data, dataset_columns)
        prediction = get_prediction_data(data, dataset_columns, column_mapping.pos_label)

        if self.probas_threshold is None and self.k is None:
            return data[dataset_columns.utility_columns.target], prediction

        if len(prediction.labels) > 2 or prediction.prediction_probas is None:
            raise ValueError("Top K / Threshold parameter can be used only with binary classification with probas")

        pos_label, neg_label = prediction.prediction_probas.columns
        threshold = self.probas_threshold

        if self.k is not None:
            threshold = k_probability_threshold(prediction.prediction_probas, self.k)

        prediction_labels = prediction.prediction_probas[pos_label].apply(
            lambda x: pos_label if x >= threshold else neg_label
        )
        return data[dataset_columns.utility_columns.target], PredictionData(
            predictions=prediction_labels,
            prediction_probas=prediction.prediction_probas,
            labels=prediction.labels,
        )



"""T test of two samples.

Name: "t_test"

Import:

    >>> from evidently.calculations.stattests import t_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import t_test
    >>> options = DataDriftOptions(all_features_stattest=t_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="t_test")
"""
from typing import Tuple

import pandas as pd
from scipy.stats import ttest_ind

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _t_test2samp(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Compute the two-sample t test between reference and current
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: level of significance (default will be 0.05)
    Returns:
        p_value: two-tailed p-value
        test_result: whether drift is detected
    """
    p_value = ttest_ind(reference_data, current_data)[1]
    return p_value, p_value < threshold


t_test = StatTest(
    name="t_test",
    display_name="t_test",
    func=_t_test2samp,
    allowed_feature_types=["num"],
)

register_stattest(t_test)



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class RegQualityMetricsBarWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [RegressionPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        results = RegressionPerformanceAnalyzer.get_results(analyzers_results)

        if (
            results.columns.utility_columns.target is None
            or results.columns.utility_columns.prediction is None
        ):
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )
            return None

        if self.dataset == "reference":
            results_metrics = results.reference_metrics

        elif self.dataset == "current":
            results_metrics = results.current_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] required '{self.dataset}' results from"
                f" {RegressionPerformanceAnalyzer.__name__} but no data found"
            )

        if results_metrics is None:
            return None

        return BaseWidgetInfo(
            title=self.title,
            type="counter",
            size=2,
            params={
                "counters": [
                    {
                        "value": f"{round(results_metrics.mean_error, 2)}"
                        f" ({round(results_metrics.error_std, 2)})",
                        "label": "ME",
                    },
                    {
                        "value": f"{round(results_metrics.mean_abs_error, 2)}"
                        f" ({round(results_metrics.abs_error_std, 2)})",
                        "label": "MAE",
                    },
                    {
                        "value": f"{round(results_metrics.mean_abs_perc_error, 2)}"
                        f" ({round(results_metrics.abs_perc_error_std, 2)})",
                        "label": "MAPE",
                    },
                ]
            },
        )



from typing import Generator

from evidently.analyzers.data_drift_analyzer import DataDriftAnalyzer
from evidently.model_monitoring.monitoring import MetricsType
from evidently.model_monitoring.monitoring import ModelMonitor
from evidently.model_monitoring.monitoring import ModelMonitoringMetric


class DataDriftMonitorMetrics:
    _tag = "data_drift"
    value = ModelMonitoringMetric(
        f"{_tag}:value", ["feature", "feature_type", "stat_test"]
    )
    dataset_drift = ModelMonitoringMetric(f"{_tag}:dataset_drift")
    share_drifted_features = ModelMonitoringMetric(f"{_tag}:share_drifted_features")
    n_drifted_features = ModelMonitoringMetric(f"{_tag}:n_drifted_features")


class DataDriftMonitor(ModelMonitor):
    def monitor_id(self) -> str:
        return "data_drift"

    def analyzers(self):
        return [DataDriftAnalyzer]

    def metrics(self, analyzer_results) -> Generator[MetricsType, None, None]:
        data_drift_results = DataDriftAnalyzer.get_results(analyzer_results)
        yield DataDriftMonitorMetrics.share_drifted_features.create(
            data_drift_results.metrics.share_of_drifted_columns
        )
        yield DataDriftMonitorMetrics.n_drifted_features.create(
            data_drift_results.metrics.number_of_drifted_columns
        )
        yield DataDriftMonitorMetrics.dataset_drift.create(
            data_drift_results.metrics.dataset_drift
        )

        for feature_name in data_drift_results.columns.get_all_features_list(
            cat_before_num=True
        ):
            feature_metric = data_drift_results.metrics.drift_by_columns[feature_name]
            yield DataDriftMonitorMetrics.value.create(
                feature_metric.drift_score,
                dict(
                    feature=feature_name,
                    feature_type=feature_metric.column_type,
                    stat_test=feature_metric.stattest_name,
                ),
            )



import copy
import dataclasses
import json
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.regression_performance import error_bias_table
from evidently.calculations.regression_performance import error_with_quantiles
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import DataDefinition


@dataclasses.dataclass
class RegressionErrorBiasTableResults:
    top_error: float
    current_plot_data: pd.DataFrame
    reference_plot_data: Optional[pd.DataFrame]
    target_name: str
    prediction_name: str
    num_feature_names: List[str]
    cat_feature_names: List[str]
    error_bias: Optional[dict] = None
    columns: Optional[List[str]] = None


class RegressionErrorBiasTable(Metric[RegressionErrorBiasTableResults]):
    # by default, we get 5% values for the error bias calculations
    TOP_ERROR_DEFAULT = 0.05
    TOP_ERROR_MIN = 0
    TOP_ERROR_MAX = 0.5
    top_error: float
    columns: Optional[List[str]]
    text_features_gen: Optional[
        Dict[str, Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]]
    ]

    def __init__(self, columns: Optional[List[str]] = None, top_error: Optional[float] = None):
        if top_error is None:
            self.top_error = self.TOP_ERROR_DEFAULT

        else:
            self.top_error = top_error

        self.columns = columns
        self.text_features_gen = None

    def required_features(self, data_definition: DataDefinition):
        if len(data_definition.get_columns("text_features")) > 0:
            text_cols = [col.column_name for col in data_definition.get_columns("text_features")]
            text_features_gen = {}
            text_features_gen_result = []
            for col in text_cols:
                col_dict: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]] = {}
                col_dict[f"{col}: Text Length"] = TextLength(col)
                col_dict[f"{col}: Non Letter Character %"] = NonLetterCharacterPercentage(col)
                col_dict[f"{col}: OOV %"] = OOVWordsPercentage(col)

                text_features_gen_result += [
                    col_dict[f"{col}: Text Length"],
                    col_dict[f"{col}: Non Letter Character %"],
                    col_dict[f"{col}: OOV %"],
                ]
                text_features_gen[col] = col_dict
            self.text_features_gen = text_features_gen

            return text_features_gen_result
        else:
            return []

    def get_parameters(self) -> tuple:
        return ()

    def calculate(self, data: InputData) -> RegressionErrorBiasTableResults:
        if self.top_error <= self.TOP_ERROR_MIN or self.top_error >= self.TOP_ERROR_MAX:
            raise ValueError(
                f"Cannot calculate error bias - "
                f"top error should be in range ({self.TOP_ERROR_MIN}, {self.TOP_ERROR_MAX})."
            )

        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        curr_df = data.current_data
        ref_df = data.reference_data

        if self.columns is None:
            columns = list(curr_df.columns)

        else:
            columns = self.columns

        if target_name is None:
            raise ValueError("Target column should be present.")

        if prediction_name is None:
            raise ValueError("Prediction column should be present.")

        if not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")

        num_feature_names = list(np.intersect1d(dataset_columns.num_feature_names, columns))
        cat_feature_names = list(np.intersect1d(dataset_columns.cat_feature_names, columns))
        # process text columns
        if (
            self.text_features_gen is not None
            and len(np.intersect1d(list(self.text_features_gen.keys()), columns)) >= 1
        ):
            for col in np.intersect1d(list(self.text_features_gen.keys()), columns):
                num_feature_names += list(self.text_features_gen[col].keys())
                columns += list(self.text_features_gen[col].keys())
                curr_text_df = pd.concat(
                    [data.get_current_column(x.feature_name()) for x in list(self.text_features_gen[col].values())],
                    axis=1,
                )
                curr_text_df.columns = list(self.text_features_gen[col].keys())
                curr_df = pd.concat([curr_df.reset_index(drop=True), curr_text_df.reset_index(drop=True)], axis=1)

                if ref_df is not None:
                    ref_text_df = pd.concat(
                        [
                            data.get_reference_column(x.feature_name())
                            for x in list(self.text_features_gen[col].values())
                        ],
                        axis=1,
                    )
                    ref_text_df.columns = list(self.text_features_gen[col].keys())
                    ref_df = pd.concat([ref_df.reset_index(drop=True), ref_text_df.reset_index(drop=True)], axis=1)
        columns_ext = np.union1d(columns, [target_name, prediction_name])
        curr_df = self._make_df_for_plot(curr_df[columns_ext], target_name, prediction_name, None)

        if ref_df is not None:
            ref_df = self._make_df_for_plot(ref_df[columns_ext], target_name, prediction_name, None)

        err_quantiles = error_with_quantiles(curr_df, prediction_name, target_name, quantile=self.top_error)

        feature_bias = error_bias_table(curr_df, err_quantiles, num_feature_names, cat_feature_names)
        error_bias = {
            feature: dict(feature_type=bias.feature_type, **bias.as_dict("current_"))
            for feature, bias in feature_bias.items()
        }

        if error_bias is not None:
            error_bias = copy.deepcopy(error_bias)

        else:
            error_bias = None

        if ref_df is not None:
            ref_err_quantiles = error_with_quantiles(ref_df, prediction_name, target_name, quantile=self.top_error)
            ref_feature_bias = error_bias_table(ref_df, ref_err_quantiles, num_feature_names, cat_feature_names)
            ref_error_bias = {
                feature: dict(feature_type=bias.feature_type, **bias.as_dict("ref_"))
                for feature, bias in ref_feature_bias.items()
            }
            if ref_error_bias is not None:
                if error_bias is not None:
                    for feature_name, reference_bias in ref_error_bias.items():
                        error_bias[feature_name].update(reference_bias)

                else:
                    error_bias = copy.deepcopy(ref_error_bias)

        if error_bias:
            error_bias_res = error_bias
        else:
            error_bias_res = {}

        columns = list(np.intersect1d(curr_df.columns, columns))

        return RegressionErrorBiasTableResults(
            top_error=self.top_error,
            current_plot_data=curr_df,
            reference_plot_data=ref_df,
            target_name=target_name,
            prediction_name=prediction_name,
            num_feature_names=[str(v) for v in num_feature_names],
            cat_feature_names=[str(v) for v in cat_feature_names],
            error_bias=error_bias_res,
            columns=[str(v) for v in columns],
        )

    @staticmethod
    def _make_df_for_plot(
        df: pd.DataFrame, target_name: str, prediction_name: str, datetime_column_name: Optional[str]
    ):
        result = df.replace([np.inf, -np.inf], np.nan)
        if datetime_column_name is not None:
            result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name, datetime_column_name])
            return result.sort_values(datetime_column_name)
        result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name])
        return result.sort_index()


@default_renderer(wrap_type=RegressionErrorBiasTable)
class RegressionErrorBiasTableRenderer(MetricRenderer):
    def render_json(self, obj: RegressionErrorBiasTable) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result.pop("current_plot_data", None)
        result.pop("reference_plot_data", None)
        return result

    def render_html(self, obj: RegressionErrorBiasTable) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_data = result.current_plot_data
        reference_data = result.reference_plot_data
        target_name = result.target_name
        prediction_name = result.prediction_name

        if reference_data is not None:
            ref_error = reference_data[prediction_name] - reference_data[target_name]
            current_error = current_data[prediction_name] - current_data[target_name]

            ref_quantile_top = np.quantile(ref_error, obj.top_error)
            ref_quantile_other = np.quantile(ref_error, 1 - obj.top_error)

            current_quantile_top = np.quantile(current_error, obj.top_error)
            current_quantile_other = np.quantile(current_error, 1 - obj.top_error)

            # create subplots
            reference_data["dataset"] = "Reference"
            reference_data["Error bias"] = list(
                map(self._error_bias_string(ref_quantile_top, ref_quantile_other), ref_error)
            )

            current_data["dataset"] = "Current"
            current_data["Error bias"] = list(
                map(self._error_bias_string(current_quantile_top, current_quantile_other), current_error)
            )
            merged_data = pd.concat([reference_data, current_data])

            reference_data.drop(["dataset", "Error bias"], axis=1, inplace=True)
            current_data.drop(["dataset", "Error bias"], axis=1, inplace=True)

            params_data = []
            additional_graphs_data = []

            for feature_name in result.num_feature_names:
                feature_type = "num"

                feature_hist = px.histogram(
                    merged_data,
                    x=feature_name,
                    color="Error bias",
                    facet_col="dataset",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={
                        "dataset": ["Reference", "Current"],
                        "Error bias": ["Underestimation", "Overestimation", "Majority"],
                    },
                )

                feature_hist_json = json.loads(feature_hist.to_json())

                segment_fig = make_subplots(rows=1, cols=2, subplot_titles=("Reference", "Current"))

                segment_fig.add_trace(
                    go.Scatter(
                        x=reference_data[target_name],
                        y=reference_data[prediction_name],
                        mode="markers",
                        marker=dict(
                            size=6,
                            cmax=max(max(reference_data[feature_name]), max(current_data[feature_name])),
                            cmin=min(min(reference_data[feature_name]), min(current_data[feature_name])),
                            color=reference_data[feature_name],
                        ),
                        showlegend=False,
                    ),
                    row=1,
                    col=1,
                )

                segment_fig.add_trace(
                    go.Scatter(
                        x=current_data[target_name],
                        y=current_data[prediction_name],
                        mode="markers",
                        marker=dict(
                            size=6,
                            cmax=max(max(reference_data[feature_name]), max(current_data[feature_name])),
                            cmin=min(min(reference_data[feature_name]), min(current_data[feature_name])),
                            color=current_data[feature_name],
                            colorbar=dict(title=feature_name),
                        ),
                        showlegend=False,
                    ),
                    row=1,
                    col=2,
                )

                # Update xaxis properties
                segment_fig.update_xaxes(title_text="Actual Value", showgrid=True, row=1, col=1)
                segment_fig.update_xaxes(title_text="Actual Value", showgrid=True, row=1, col=2)

                # Update yaxis properties
                segment_fig.update_yaxes(title_text="Predicted Value", showgrid=True, row=1, col=1)
                segment_fig.update_yaxes(title_text="Predicted Value", showgrid=True, row=1, col=2)

                segment_json = json.loads(segment_fig.to_json())

                if result.error_bias is None:
                    raise ValueError("RegressionErrorBiasTableRenderer got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {"title": "Predicted vs Actual", "id": feature_name + "_segm"},
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": round(result.error_bias[feature_name]["ref_majority"], 2),
                        "f4": round(result.error_bias[feature_name]["ref_under"], 2),
                        "f5": round(result.error_bias[feature_name]["ref_over"], 2),
                        "f6": round(result.error_bias[feature_name]["ref_range"], 2),
                        "f7": round(result.error_bias[feature_name]["current_majority"], 2),
                        "f8": round(result.error_bias[feature_name]["current_under"], 2),
                        "f9": round(result.error_bias[feature_name]["current_over"], 2),
                        "f10": round(result.error_bias[feature_name]["current_range"], 2),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist",
                        {"data": feature_hist_json["data"], "layout": feature_hist_json["layout"]},
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm", {"data": segment_json["data"], "layout": segment_json["layout"]}
                    )
                )

            for feature_name in result.cat_feature_names:
                feature_type = "cat"

                feature_hist = px.histogram(
                    merged_data,
                    x=feature_name,
                    color="Error bias",
                    facet_col="dataset",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={
                        "dataset": ["Reference", "Current"],
                        "Error bias": ["Underestimation", "Overestimation", "Majority"],
                    },
                )

                feature_hist_json = json.loads(feature_hist.to_json())
                segment_fig = px.scatter(
                    merged_data[~merged_data[feature_name].isna()],
                    x=target_name,
                    y=prediction_name,
                    color=feature_name,
                    facet_col="dataset",
                )

                segment_json = json.loads(segment_fig.to_json())

                if result.error_bias is None:
                    raise ValueError("RegressionErrorBiasTableRenderer got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {"title": "Predicted vs Actual", "id": feature_name + "_segm"},
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": str(result.error_bias[feature_name]["ref_majority"]),
                        "f4": str(result.error_bias[feature_name]["ref_under"]),
                        "f5": str(result.error_bias[feature_name]["ref_over"]),
                        "f6": str(result.error_bias[feature_name]["ref_range"]),
                        "f7": str(result.error_bias[feature_name]["current_majority"]),
                        "f8": str(result.error_bias[feature_name]["current_under"]),
                        "f9": str(result.error_bias[feature_name]["current_over"]),
                        "f10": int(result.error_bias[feature_name]["current_range"]),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist",
                        {"data": feature_hist_json["data"], "layout": feature_hist_json["layout"]},
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm", {"data": segment_json["data"], "layout": segment_json["layout"]}
                    )
                )
            if result.columns is None:
                size = 0
            else:
                size = len(result.columns)
            widget_info = BaseWidgetInfo(
                title="",
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": min(size, 10),
                    "columns": [
                        {"title": "Feature", "field": "f1"},
                        {"title": "Type", "field": "f2"},
                        {"title": "REF: Majority", "field": "f3"},
                        {"title": "REF: Under", "field": "f4"},
                        {"title": "REF: Over", "field": "f5"},
                        {"title": "REF: Range(%)", "field": "f6"},
                        {"title": "CURR: Majority", "field": "f7"},
                        {"title": "CURR: Under", "field": "f8"},
                        {"title": "CURR: Over", "field": "f9"},
                        {"title": "CURR: Range(%)", "field": "f10", "sort": "desc"},
                    ],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )

        else:
            error = current_data[prediction_name] - current_data[target_name]

            quantile_top = np.quantile(error, obj.top_error)
            quantile_other = np.quantile(error, 1 - obj.top_error)

            current_data["Error bias"] = list(
                map(
                    lambda x: "Underestimation"
                    if x <= quantile_top
                    else "Majority"
                    if x < quantile_other
                    else "Overestimation",
                    error,
                )
            )

            params_data = []
            additional_graphs_data = []

            for feature_name in result.num_feature_names:  # + cat_feature_names: #feature_names:

                feature_type = "num"

                hist = px.histogram(
                    current_data,
                    x=feature_name,
                    color="Error bias",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={"Error bias": ["Underestimation", "Overestimation", "Majority"]},
                )

                hist_figure = json.loads(hist.to_json())

                segm = px.scatter(current_data, x=target_name, y=prediction_name, color=feature_name)
                segm_figure = json.loads(segm.to_json())

                if result.error_bias is None:
                    raise ValueError("Widget RegressionErrorBiasTableRenderer got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {"title": "Predicted vs Actual", "id": feature_name + "_segm"},
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": round(result.error_bias[feature_name]["current_majority"], 2),
                        "f4": round(result.error_bias[feature_name]["current_under"], 2),
                        "f5": round(result.error_bias[feature_name]["current_over"], 2),
                        "f6": round(result.error_bias[feature_name]["current_range"], 2),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist", {"data": hist_figure["data"], "layout": hist_figure["layout"]}
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm", {"data": segm_figure["data"], "layout": segm_figure["layout"]}
                    )
                )

            for feature_name in result.cat_feature_names:  # feature_names:

                feature_type = "cat"

                hist = px.histogram(
                    current_data,
                    x=feature_name,
                    color="Error bias",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={"Error bias": ["Underestimation", "Overestimation", "Majority"]},
                )

                hist_figure = json.loads(hist.to_json())

                initial_type = current_data[feature_name].dtype
                current_data[feature_name] = current_data[feature_name].astype(str)
                segm = px.scatter(current_data, x=target_name, y=prediction_name, color=feature_name)
                current_data[feature_name] = current_data[feature_name].astype(initial_type)

                segm_figure = json.loads(segm.to_json())

                if result.error_bias is None:
                    raise ValueError("RegressionErrorBiasTableRenderer got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {"title": "Predicted vs Actual", "id": feature_name + "_segm"},
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": str(result.error_bias[feature_name]["current_majority"]),
                        "f4": str(result.error_bias[feature_name]["current_under"]),
                        "f5": str(result.error_bias[feature_name]["current_over"]),
                        "f6": str(result.error_bias[feature_name]["current_range"]),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist", {"data": hist_figure["data"], "layout": hist_figure["layout"]}
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm", {"data": segm_figure["data"], "layout": segm_figure["layout"]}
                    )
                )

            current_data.drop("Error bias", axis=1, inplace=True)

            if result.columns is None:
                size = 0
            else:
                size = len(result.columns)

            widget_info = BaseWidgetInfo(
                title="",
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": min(size, 10),
                    "columns": [
                        {"title": "Feature", "field": "f1"},
                        {"title": "Type", "field": "f2"},
                        {"title": "Majority", "field": "f3"},
                        {"title": "Underestimation", "field": "f4"},
                        {"title": "Overestimation", "field": "f5"},
                        {"title": "Range(%)", "field": "f6", "sort": "desc"},
                    ],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )
        top_error_percents = int(result.top_error * 100)

        return [
            header_text(
                label=f"Error Bias: Mean/Most Common Feature Value per Group (Top - {top_error_percents}% Errors)"
            ),
            widget_info,
        ]

    @staticmethod
    def _error_bias_string(quantile_top, quantile_other):
        def __error_bias_string(error):
            if error <= quantile_top:
                return "Underestimation"

            if error < quantile_other:
                return "Majority"

            return "Overestimation"

        return __error_bias_string



from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

from evidently.calculations.stattests import PossibleStatTestType


def resolve_stattest_threshold(
    feature_name: str,
    feature_type: str,
    stattest: Optional[PossibleStatTestType],
    cat_stattest: Optional[PossibleStatTestType],
    num_stattest: Optional[PossibleStatTestType],
    text_stattest: Optional[PossibleStatTestType],
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]],
    stattest_threshold: Optional[float],
    cat_stattest_threshold: Optional[float],
    num_stattest_threshold: Optional[float],
    text_stattest_threshold: Optional[float],
    per_column_stattest_threshold: Optional[Dict[str, float]],
) -> Tuple[Optional[PossibleStatTestType], Optional[float]]:
    return (
        _calculate_stattest(
            feature_name,
            feature_type,
            stattest,
            cat_stattest,
            num_stattest,
            text_stattest,
            per_column_stattest,
        ),
        _calculate_threshold(
            feature_name,
            feature_type,
            stattest_threshold,
            cat_stattest_threshold,
            num_stattest_threshold,
            text_stattest_threshold,
            per_column_stattest_threshold,
        ),
    )


def _calculate_stattest(
    feature_name: str,
    feature_type: str,
    stattest: Optional[PossibleStatTestType] = None,
    cat_stattest: Optional[PossibleStatTestType] = None,
    num_stattest: Optional[PossibleStatTestType] = None,
    text_stattest: Optional[PossibleStatTestType] = None,
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
) -> Optional[PossibleStatTestType]:
    func = None if stattest is None else stattest
    if feature_type == "cat":
        type_func = cat_stattest
    elif feature_type == "num":
        type_func = num_stattest
    elif feature_type == "text":
        type_func = text_stattest
    else:
        raise ValueError(f"Unexpected feature type {feature_type}.")
    func = func if type_func is None else type_func
    if per_column_stattest is None:
        return func
    return per_column_stattest.get(feature_name, func)


def _calculate_threshold(
    feature_name: str,
    feature_type: str,
    stattest_threshold: Optional[float] = None,
    cat_stattest_threshold: Optional[float] = None,
    num_stattest_threshold: Optional[float] = None,
    text_stattest_threshold: Optional[float] = None,
    per_column_stattest_threshold: Optional[Dict[str, float]] = None,
) -> Optional[float]:
    if per_column_stattest_threshold is not None:
        return per_column_stattest_threshold.get(feature_name)

    if cat_stattest_threshold is not None and feature_type == "cat":
        return cat_stattest_threshold

    if num_stattest_threshold is not None and feature_type == "num":
        return num_stattest_threshold

    if text_stattest_threshold is not None and feature_type == "text":
        return text_stattest_threshold

    if stattest_threshold is not None:
        return stattest_threshold
    return None


def roc_auc_domain_classifier(X_train, X_test, y_train, y_test) -> Tuple:
    pipeline = Pipeline(
        [
            ("vectorization", TfidfVectorizer(sublinear_tf=True, max_df=0.5, stop_words="english")),
            ("classification", SGDClassifier(alpha=0.0001, max_iter=50, penalty="l1", loss="modified_huber")),
        ]
    )

    pipeline.fit(X_train, y_train)
    y_pred_proba = pipeline.predict_proba(X_test)[:, 1]
    roc_auc = roc_auc_score(y_test, y_pred_proba)
    return roc_auc, y_pred_proba, pipeline


def roc_auc_random_classifier_percentile(y_test: np.array, p_value=0.05, iter_num=1000, seed=42) -> float:
    def calc_roc_auc_random(y_test, seed=None):
        np.random.seed(seed)
        y_random_pred = np.random.rand(len(y_test))
        roc_auc_random = roc_auc_score(y_test, y_random_pred)
        return roc_auc_random

    np.random.seed(seed)
    seeds = np.random.randint(0, iter_num * 10, size=iter_num)
    roc_auc_values = [calc_roc_auc_random(y_test, seed=seed) for seed in seeds]
    return np.percentile(roc_auc_values, 100 * (1 - p_value))


def calculate_text_drift_score(reference_data: pd.Series, current_data: pd.Series, p_value=0.05) -> Tuple[float, bool]:
    domain_data = pd.concat(
        [
            pd.DataFrame({"text": reference_data, "target": 0}),
            pd.DataFrame({"text": current_data, "target": 1}),
        ]
    )

    X_train, X_test, y_train, y_test = train_test_split(
        domain_data["text"],
        domain_data["target"],
        test_size=0.5,
        random_state=42,
        shuffle=True,
    )

    domain_classifier_roc_auc, _, _ = roc_auc_domain_classifier(
        X_train,
        X_test,
        y_train,
        y_test,
    )

    random_classifier_95_percentile = roc_auc_random_classifier_percentile(y_test, p_value=p_value)
    return domain_classifier_roc_auc, domain_classifier_roc_auc > random_classifier_95_percentile


def get_typical_examples(X_test, y_test, y_pred_proba, examples_num=10) -> Tuple[List[str], List[str]]:
    typical_examples = pd.DataFrame({"text": X_test, "label": y_test, "predict_proba": y_pred_proba})

    typical_current = typical_examples[typical_examples["label"] == 1]
    typical_current.sort_values("predict_proba", ascending=False, inplace=True)
    typical_current = typical_current.head(examples_num)

    typical_reference = typical_examples[typical_examples["label"] == 0]
    typical_reference.sort_values("predict_proba", ascending=True, inplace=True)
    typical_reference = typical_reference.head(examples_num)
    return list(typical_current["text"]), list(typical_reference["text"])


def get_typical_words(pipeline, words_num=25) -> Tuple[List[str], List[str]]:
    weights_df = pd.DataFrame({"weight": pipeline["classification"].coef_[0]})
    weights_df.reset_index(inplace=True)
    weights_df.rename(columns={"index": "feature_ind"}, inplace=True)
    weights_df = weights_df[weights_df["weight"] != 0]

    # build inverted index for vocabulary
    inverted_vocabulary = {value: key for key, value in pipeline["vectorization"].vocabulary_.items()}
    words_typical_current = weights_df[weights_df["weight"] > 0]
    words_typical_current.sort_values("weight", ascending=False, inplace=True)
    words_typical_current = words_typical_current.head(words_num)
    words_typical_current["word"] = words_typical_current["feature_ind"].apply(lambda x: inverted_vocabulary[x])

    words_typical_reference = weights_df[weights_df["weight"] < 0]
    words_typical_reference.sort_values("weight", ascending=True, inplace=True)
    words_typical_reference = words_typical_reference.head(words_num)
    words_typical_reference["word"] = words_typical_reference["feature_ind"].apply(lambda x: inverted_vocabulary[x])

    return list(words_typical_current["word"]), list(words_typical_reference["word"])


def get_text_data_for_plots(reference_data: pd.Series, current_data: pd.Series):
    domain_data = pd.concat(
        [
            pd.DataFrame({"text": reference_data, "target": 0}),
            pd.DataFrame({"text": current_data, "target": 1}),
        ]
    )

    X_train, X_test, y_train, y_test = train_test_split(
        domain_data["text"],
        domain_data["target"],
        test_size=0.5,
        random_state=42,
        shuffle=True,
    )

    _, y_pred_proba, classifier_pipeline = roc_auc_domain_classifier(
        X_train,
        X_test,
        y_train,
        y_test,
    )
    # get examples more characteristic of current or reference dataset
    typical_examples_cur, typical_examples_ref = get_typical_examples(X_test, y_test, y_pred_proba)

    # get words more characteristic of current or reference dataset
    typical_words_cur, typical_words_ref = get_typical_words(classifier_pipeline)

    return typical_examples_cur, typical_examples_ref, typical_words_cur, typical_words_ref



import dataclasses
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_absolute_percentage_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.regression_performance import calculate_regression_performance
from evidently.metrics.utils import apply_func_to_binned_data
from evidently.metrics.utils import make_target_bins_for_reg_plots
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import make_hist_for_cat_plot
from evidently.utils.visualizations import make_hist_for_num_plot


@dataclasses.dataclass
class RegressionQualityMetricResults:
    columns: DatasetColumns
    r2_score: float
    rmse: float
    rmse_default: float
    mean_error: float
    me_default_sigma: float
    me_hist_for_plot: Dict[str, pd.Series]
    mean_abs_error: float
    mean_abs_error_default: float
    mean_abs_perc_error: float
    mean_abs_perc_error_default: float
    abs_error_max: float
    abs_error_max_default: float
    error_std: float
    abs_error_std: float
    abs_perc_error_std: float
    error_normality: dict
    underperformance: dict
    hist_for_plot: Dict[str, pd.Series]
    vals_for_plots: Dict[str, Dict[str, pd.Series]]
    error_bias: Optional[dict] = None
    mean_error_ref: Optional[float] = None
    mean_abs_error_ref: Optional[float] = None
    mean_abs_perc_error_ref: Optional[float] = None
    rmse_ref: Optional[float] = None
    r2_score_ref: Optional[float] = None
    abs_error_max_ref: Optional[float] = None
    underperformance_ref: Optional[dict] = None
    error_std_ref: Optional[float] = None
    abs_error_std_ref: Optional[float] = None
    abs_perc_error_std_ref: Optional[float] = None


class RegressionQualityMetric(Metric[RegressionQualityMetricResults]):
    def calculate(self, data: InputData) -> RegressionQualityMetricResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction

        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        if not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")
        current_metrics = calculate_regression_performance(
            dataset=data.current_data, columns=dataset_columns, error_bias_prefix="current_"
        )
        error_bias = current_metrics.error_bias
        reference_metrics = None

        if data.reference_data is not None:
            ref_columns = process_columns(data.reference_data, data.column_mapping)
            reference_metrics = calculate_regression_performance(
                dataset=data.reference_data, columns=ref_columns, error_bias_prefix="ref_"
            )

            if reference_metrics is not None and reference_metrics.error_bias:
                for feature_name, current_bias in reference_metrics.error_bias.items():
                    if feature_name in error_bias:
                        error_bias[feature_name].update(current_bias)

                    else:
                        error_bias[feature_name] = current_bias

        r2_score_value = r2_score(
            y_true=data.current_data[target_name],
            y_pred=data.current_data[prediction_name],
        )
        rmse_score_value = mean_squared_error(
            y_true=data.current_data[target_name],
            y_pred=data.current_data[prediction_name],
        )

        # mae default values
        dummy_preds = data.current_data[target_name].median()
        mean_abs_error_default = mean_absolute_error(
            y_true=data.current_data[target_name], y_pred=[dummy_preds] * data.current_data.shape[0]
        )
        # rmse default values
        rmse_ref = None
        if data.reference_data is not None:
            rmse_ref = mean_squared_error(
                y_true=data.reference_data[target_name],
                y_pred=data.reference_data[prediction_name],
            )
        dummy_preds = data.current_data[target_name].mean()
        rmse_default = mean_squared_error(
            y_true=data.current_data[target_name], y_pred=[dummy_preds] * data.current_data.shape[0]
        )
        # mape default values
        # optimal constant for mape
        s = data.current_data[target_name]
        inv_y = 1 / s[s != 0].values
        w = inv_y / sum(inv_y)
        idxs = np.argsort(w)
        sorted_w = w[idxs]
        sorted_w_cumsum = np.cumsum(sorted_w)
        idx = np.where(sorted_w_cumsum > 0.5)[0][0]
        pos = idxs[idx]
        dummy_preds = s[s != 0].values[pos]

        mean_abs_perc_error_default = (
            mean_absolute_percentage_error(
                y_true=data.current_data[target_name],
                y_pred=[dummy_preds] * data.current_data.shape[0],
            )
            * 100
        )
        #  r2_score default values
        r2_score_ref = None
        if data.reference_data is not None:
            r2_score_ref = r2_score(
                y_true=data.reference_data[target_name],
                y_pred=data.reference_data[prediction_name],
            )
        # max error default values
        abs_error_max_ref = None
        mean_error_ref = None

        if reference_metrics is not None:
            abs_error_max_ref = reference_metrics.abs_error_max
            mean_error_ref = reference_metrics.mean_error

        y_true = data.current_data[target_name]
        y_pred = data.current_data[prediction_name]
        abs_error_max_default = np.abs(y_true - y_true.median()).max()

        #  me default values
        me_default_sigma = (y_pred - y_true).std()

        # visualisation

        df_target_binned = make_target_bins_for_reg_plots(
            data.current_data, target_name, prediction_name, data.reference_data
        )
        curr_target_bins = df_target_binned.loc[df_target_binned.data == "curr", "target_binned"]
        ref_target_bins = None
        if data.reference_data is not None:
            ref_target_bins = df_target_binned.loc[df_target_binned.data == "ref", "target_binned"]
        hist_for_plot = make_hist_for_cat_plot(curr_target_bins, ref_target_bins)

        vals_for_plots = {}

        if data.reference_data is not None:
            is_ref_data = True

        else:
            is_ref_data = False

        for name, func in zip(
            ["r2_score", "rmse", "mean_abs_error", "mean_abs_perc_error"],
            [r2_score, mean_squared_error, mean_absolute_error, mean_absolute_percentage_error],
        ):
            vals_for_plots[name] = apply_func_to_binned_data(
                df_target_binned, func, target_name, prediction_name, is_ref_data
            )

        # me plot
        err_curr = data.current_data[prediction_name] - data.current_data[target_name]
        err_ref = None

        if is_ref_data:
            err_ref = data.reference_data[prediction_name] - data.reference_data[target_name]
        me_hist_for_plot = make_hist_for_num_plot(err_curr, err_ref)

        if r2_score_ref is not None:
            r2_score_ref = float(r2_score_ref)

        if rmse_ref is not None:
            rmse_ref = float(rmse_ref)

        underperformance_ref = None

        if reference_metrics is not None:
            underperformance_ref = reference_metrics.underperformance

        return RegressionQualityMetricResults(
            columns=dataset_columns,
            r2_score=r2_score_value,
            rmse=rmse_score_value,
            rmse_default=rmse_default,
            mean_error=current_metrics.mean_error,
            mean_error_ref=mean_error_ref,
            me_default_sigma=me_default_sigma,
            me_hist_for_plot=me_hist_for_plot,
            mean_abs_error=current_metrics.mean_abs_error,
            mean_abs_error_default=mean_abs_error_default,
            mean_abs_perc_error=current_metrics.mean_abs_perc_error,
            mean_abs_perc_error_default=mean_abs_perc_error_default,
            abs_error_max=current_metrics.abs_error_max,
            abs_error_max_default=abs_error_max_default,
            error_std=current_metrics.error_std,
            abs_error_std=current_metrics.abs_error_std,
            abs_perc_error_std=current_metrics.abs_perc_error_std,
            error_normality=current_metrics.error_normality,
            underperformance=current_metrics.underperformance,
            underperformance_ref=underperformance_ref,
            hist_for_plot=hist_for_plot,
            vals_for_plots=vals_for_plots,
            error_bias=error_bias,
            mean_abs_error_ref=reference_metrics.mean_abs_error if reference_metrics is not None else None,
            mean_abs_perc_error_ref=reference_metrics.mean_abs_perc_error if reference_metrics is not None else None,
            error_std_ref=reference_metrics.error_std if reference_metrics is not None else None,
            abs_error_std_ref=reference_metrics.abs_error_std if reference_metrics is not None else None,
            abs_perc_error_std_ref=reference_metrics.abs_perc_error_std if reference_metrics is not None else None,
            rmse_ref=rmse_ref,
            r2_score_ref=r2_score_ref,
            abs_error_max_ref=abs_error_max_ref,
        )


@default_renderer(wrap_type=RegressionQualityMetric)
class RegressionQualityMetricRenderer(MetricRenderer):
    def render_json(self, obj: RegressionQualityMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        # remove values with DataFrames or Series
        result.pop("hist_for_plot")
        result.pop("vals_for_plots")
        result.pop("me_hist_for_plot")
        return result

    def render_html(self, obj: RegressionQualityMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        target_name = metric_result.columns.utility_columns.target
        result = [
            header_text(label=f"Regression Model Performance. Target: '{target_name}’"),
            counter(
                title="Current: Model Quality (+/- std)",
                counters=[
                    CounterData("ME", f"{round(metric_result.mean_error, 2)} ({round(metric_result.error_std, 2)})"),
                    CounterData(
                        "MAE", f"{round(metric_result.mean_abs_error, 2)} ({round(metric_result.abs_error_std, 2)})"
                    ),
                    CounterData(
                        "MAPE",
                        (
                            f"{round(metric_result.mean_abs_perc_error, 2)}"
                            f" ({round(metric_result.abs_perc_error_std, 2)})"
                        ),
                    ),
                ],
            ),
        ]
        if (
            metric_result.mean_error_ref is not None
            and metric_result.error_std_ref is not None
            and metric_result.mean_abs_error_ref is not None
            and metric_result.abs_error_std_ref is not None
            and metric_result.mean_abs_perc_error_ref is not None
            and metric_result.abs_perc_error_std_ref is not None
        ):
            result.append(
                counter(
                    title="Reference: Model Quality (+/- std)",
                    counters=[
                        CounterData(
                            "ME", f"{round(metric_result.mean_error_ref, 2)} ({round(metric_result.error_std_ref, 2)})"
                        ),
                        CounterData(
                            "MAE",
                            (
                                f"{round(metric_result.mean_abs_error_ref, 2)}"
                                f" ({round(metric_result.abs_error_std_ref, 2)})"
                            ),
                        ),
                        CounterData(
                            "MAPE",
                            (
                                f"{round(metric_result.mean_abs_perc_error_ref, 2)}"
                                f" ({round(metric_result.abs_perc_error_std_ref, 2)})"
                            ),
                        ),
                    ],
                ),
            )
        return result



from typing import Optional

import plotly.graph_objs as go

from evidently.options import ColorOptions
from evidently.utils.visualizations import Distribution


def plot_distr(*, hist_curr, hist_ref=None, orientation="v", color_options: ColorOptions):
    fig = go.Figure()

    fig.add_trace(
        go.Bar(
            name="current",
            x=hist_curr["x"],
            y=hist_curr["count"],
            marker_color=color_options.get_current_data_color(),
            orientation=orientation,
        )
    )
    if hist_ref is not None:
        fig.add_trace(
            go.Bar(
                name="reference",
                x=hist_ref["x"],
                y=hist_ref["count"],
                marker_color=color_options.get_reference_data_color(),
                orientation=orientation,
            )
        )

    return fig


def get_distribution_plot_figure(
    *,
    current_distribution: Distribution,
    reference_distribution: Optional[Distribution],
    color_options: ColorOptions,
    orientation: str = "v",
) -> go.Figure:
    fig = go.Figure()
    fig.add_trace(
        go.Bar(
            name="current",
            x=current_distribution.x,
            y=current_distribution.y,
            marker_color=color_options.get_current_data_color(),
            orientation=orientation,
        )
    )
    if reference_distribution is not None:
        fig.add_trace(
            go.Bar(
                name="reference",
                x=reference_distribution.x,
                y=reference_distribution.y,
                marker_color=color_options.get_reference_data_color(),
                orientation=orientation,
            )
        )

    return fig



from typing import Dict
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.calculations.stattests import PossibleStatTestType
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.metrics import DataDriftTable
from evidently.metrics import DatasetDriftMetric
from evidently.utils.data_operations import DatasetColumns


class DataDriftPreset(MetricPreset):
    """Metric Preset for Data Drift analysis.

    Contains metrics:
    - DatasetDriftMetric
    - DataDriftTable
    """

    columns: Optional[List[str]]
    drift_share: float
    stattest: Optional[PossibleStatTestType]
    cat_stattest: Optional[PossibleStatTestType]
    num_stattest: Optional[PossibleStatTestType]
    text_stattest: Optional[PossibleStatTestType]
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]]
    stattest_threshold: Optional[float]
    cat_stattest_threshold: Optional[float]
    num_stattest_threshold: Optional[float]
    text_stattest_threshold: Optional[float]
    per_column_stattest_threshold: Optional[Dict[str, float]]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        drift_share: float = 0.5,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        super().__init__()
        self.columns = columns
        self.drift_share = drift_share
        self.stattest = stattest
        self.cat_stattest = cat_stattest
        self.num_stattest = num_stattest
        self.text_stattest = text_stattest
        self.per_column_stattest = per_column_stattest
        self.stattest_threshold = stattest_threshold
        self.cat_stattest_threshold = cat_stattest_threshold
        self.num_stattest_threshold = num_stattest_threshold
        self.text_stattest_threshold = text_stattest_threshold
        self.per_column_stattest_threshold = per_column_stattest_threshold

    def generate_metrics(self, data: InputData, columns: DatasetColumns):
        return [
            DatasetDriftMetric(
                columns=self.columns,
                drift_share=self.drift_share,
                stattest=self.stattest,
                cat_stattest=self.cat_stattest,
                num_stattest=self.num_stattest,
                text_stattest=self.text_stattest,
                per_column_stattest=self.per_column_stattest,
                stattest_threshold=self.stattest_threshold,
                cat_stattest_threshold=self.cat_stattest_threshold,
                num_stattest_threshold=self.num_stattest_threshold,
                text_stattest_threshold=self.text_stattest_threshold,
                per_column_stattest_threshold=self.per_column_stattest_threshold,
            ),
            DataDriftTable(
                columns=self.columns,
                stattest=self.stattest,
                cat_stattest=self.cat_stattest,
                num_stattest=self.num_stattest,
                text_stattest=self.text_stattest,
                per_column_stattest=self.per_column_stattest,
                stattest_threshold=self.stattest_threshold,
                cat_stattest_threshold=self.cat_stattest_threshold,
                num_stattest_threshold=self.num_stattest_threshold,
                text_stattest_threshold=self.text_stattest_threshold,
                per_column_stattest_threshold=self.per_column_stattest_threshold,
            ),
        ]



"""Additional types, classes, dataclasses, etc."""

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

# type for numeric because of mypy bug https://github.com/python/mypy/issues/3186
Numeric = Union[float, int]

# type for distributions - list of tuples (value, count)
ColumnDistribution = Dict[Any, Numeric]


class ApproxValue:
    """Class for approximate scalar value calculations"""

    DEFAULT_RELATIVE = 1e-6
    DEFAULT_ABSOLUTE = 1e-12
    value: Numeric
    _relative: Numeric
    _absolute: Numeric

    def __init__(self, value: Numeric, relative: Optional[Numeric] = None, absolute: Optional[Numeric] = None):
        self.value = value

        if relative is not None and relative <= 0:
            raise ValueError("Relative value for approx should be greater than 0")

        if relative is None:
            self._relative = self.DEFAULT_RELATIVE

        else:
            self._relative = relative

        if absolute is None:
            self._absolute = self.DEFAULT_ABSOLUTE

        else:
            self._absolute = absolute

    @property
    def tolerance(self) -> Numeric:
        relative_value = abs(self.value) * self._relative
        return max(relative_value, self._absolute)

    def __format__(self, format_spec):
        return f"{format(self.value, format_spec)} ± {format(self.tolerance, format_spec)}"

    def __repr__(self):
        return f"{self.value} ± {self.tolerance}"

    def __eq__(self, other):
        tolerance = self.tolerance
        return (self.value - tolerance) <= other <= (self.value + tolerance)

    def __lt__(self, other):
        return self.value + self.tolerance < other

    def __le__(self, other):
        return self.value - self.tolerance <= other

    def __gt__(self, other):
        return self.value - self.tolerance > other

    def __ge__(self, other):
        return self.value + self.tolerance >= other

    def as_dict(self) -> dict:
        result = {"value": self.value}

        if self._relative is not None:
            result["relative"] = self._relative

        if self._absolute is not None:
            result["absolute"] = self._absolute

        return result



#!/usr/bin/env python
# coding: utf-8
from dataclasses import dataclass
from typing import Dict
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.calculations.data_quality import DataQualityStats
from evidently.calculations.data_quality import calculate_correlations
from evidently.calculations.data_quality import calculate_data_quality_stats
from evidently.utils.data_operations import process_columns
from evidently.utils.data_operations import recognize_task


@dataclass
class DataQualityAnalyzerResults(BaseAnalyzerResult):
    """Class for all results of data quality calculations"""

    reference_features_stats: DataQualityStats
    reference_correlations: Dict[str, pd.DataFrame]
    current_features_stats: Optional[DataQualityStats] = None
    current_correlations: Optional[Dict[str, pd.DataFrame]] = None


class DataQualityAnalyzer(Analyzer):
    """Data quality analyzer
    provides detailed feature statistics and feature behavior overview
    """

    @staticmethod
    def get_results(analyzer_results) -> DataQualityAnalyzerResults:
        return analyzer_results[DataQualityAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
    ) -> DataQualityAnalyzerResults:
        """Calculates base statistics for numerical, categorical and datetime features.
        For categorical features, calculates the Cramer's v correlation matrix.
        For numerical features, Evidently calculates the Pearson, Spearman and Kendall matrices.

        Args:
            reference_data: usually the data which you used in training.
            current_data: new, unseen data to which we compare the reference data.
            column_mapping: a `ColumnMapping` object that contains references to the name of target and prediction
                columns
        Returns:
            A dictionary that contains:
                - some meta information
                - data quality metrics for all features
                - correlation matrices
        """

        columns = process_columns(reference_data, column_mapping)
        target_name = columns.utility_columns.target
        task: Optional[str]

        if column_mapping.task is not None:
            task = column_mapping.task

        elif column_mapping.task is None and target_name:
            task = recognize_task(target_name, reference_data)

        else:
            task = None

        reference_features_stats = calculate_data_quality_stats(reference_data, columns, task)

        current_features_stats: Optional[DataQualityStats]

        if current_data is not None:
            current_features_stats = calculate_data_quality_stats(current_data, columns, task)

            all_cat_features = {}

            if current_features_stats.cat_features_stats is not None:
                all_cat_features.update(current_features_stats.cat_features_stats)

            if task == "classification" and current_features_stats.target_stats is not None:
                all_cat_features.update(current_features_stats.target_stats)

            if current_features_stats.cat_features_stats is not None:
                # calculate additional stats of representation reference dataset values in the current dataset
                for feature_name, cat_feature_stats in all_cat_features.items():
                    current_values_set = set(current_data[feature_name].unique())

                    if feature_name in reference_data:
                        reference_values_set = set(reference_data[feature_name].unique())

                    else:
                        reference_values_set = set()

                    unique_in_current = current_values_set - reference_values_set
                    new_in_current_values_count: int = len(unique_in_current)
                    unique_in_reference = reference_values_set - current_values_set
                    unused_in_current_values_count: int = len(unique_in_reference)

                    # take into account that NaN values in Python sets do not support substitution correctly
                    # {nan} - {nan} can be equals {nan}
                    # use pd.isnull because it supports strings values correctly, np.isnan raises and exception
                    if any(pd.isnull(list(unique_in_current))) and any(pd.isnull(list(unique_in_reference))):
                        new_in_current_values_count -= 1
                        unused_in_current_values_count -= 1

                    cat_feature_stats.new_in_current_values_count = new_in_current_values_count
                    cat_feature_stats.unused_in_current_values_count = unused_in_current_values_count

        else:
            current_features_stats = None

        # calculate correlations
        reference_correlations: Dict = calculate_correlations(reference_data, columns)

        if current_features_stats is not None:
            current_correlations: Dict = calculate_correlations(current_data, columns)

        else:
            current_correlations = {}

        results = DataQualityAnalyzerResults(
            columns=columns,
            reference_features_stats=reference_features_stats,
            reference_correlations=reference_correlations,
        )

        if current_features_stats is not None:
            results.current_features_stats = current_features_stats
            results.current_correlations = current_correlations

        return results



import dataclasses
from typing import Dict
from typing import List

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import make_hist_for_cat_plot
from evidently.utils.visualizations import plot_distr_subplots


@dataclasses.dataclass
class ClassificationClassBalanceResult:
    plot_data: Dict[str, int]


class ClassificationClassBalance(Metric[ClassificationClassBalanceResult]):
    def calculate(self, data: InputData) -> ClassificationClassBalanceResult:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        ref_target = None
        if data.reference_data is not None:
            ref_target = data.reference_data[target_name]
        plot_data = make_hist_for_cat_plot(data.current_data[target_name], ref_target)

        return ClassificationClassBalanceResult(plot_data=plot_data)


@default_renderer(wrap_type=ClassificationClassBalance)
class ClassificationClassBalanceRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationClassBalance) -> dict:
        return {}

    def render_html(self, obj: ClassificationClassBalance) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        current_plot_data = metric_result.plot_data["current"]
        reference_plot_data = None
        if "reference" in metric_result.plot_data.keys():
            reference_plot_data = metric_result.plot_data["reference"]

        fig = plot_distr_subplots(
            hist_curr=current_plot_data,
            hist_ref=reference_plot_data,
            xaxis_name="Class",
            yaxis_name="Number Of Objects",
            same_color=True,
            color_options=self.color_options,
        )
        return [
            header_text(label="Class Representation"),
            BaseWidgetInfo(
                title="",
                size=2,
                type="big_graph",
                params={"data": fig["data"], "layout": fig["layout"]},
            ),
        ]
